
""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject
from dataclay.util.management.classmgr.Type import Type


class PropertyStubInfo(ManagementObject):
    _fields = ["namespace",
               "propertyName",
               "namespaceID",
               "propertyID",
               "propertyType",
               "getterOperationID",
               "setterOperationID",
               "beforeUpdate",
               "afterUpdate",
               "inMaster"
               ]

    _internal_fields = list()

    _typed_fields = {"propertyType": Type}




""" Class description goes here. """
import sys
from weakref import WeakValueDictionary

"""
Created on 26 ene. 2018

@author: dgasull
"""
import logging
import time
import gc
import traceback

try:
    import tracemalloc
except ImportError:
    tracemalloc = None

import psutil

from dataclay.heap.HeapManager import HeapManager
from dataclay.serialization.lib.SerializationLibUtils import SerializationLibUtilsSingleton
from dataclay.serialization.lib.DeserializationLibUtils import DeserializationLibUtilsSingleton
from dataclay.DataClayObjProperties import DCLAY_PROPERTY_PREFIX
from dataclay.util import Configuration


class ExecutionEnvironmentHeapManager(HeapManager):

    """
    @summary: This class is intended to manage all dataClay objects in EE runtime's memory.
    """

    def __init__(self, theruntime):
        """
        @postcondition: Constructor of the object 
        @param theruntime: Runtime being managed 
        """ 
        HeapManager.__init__(self, theruntime)  
        """ During a flush of all objects in Heap, if GC is being processed, wait, and check after time specified here in seconds """
        self.TIME_WAIT_FOR_GC_TO_FINISH = 1
        
        """ Execution Environment being managed """
        self.exec_env = None 
        
        # Retained objects so they cannot be GC by PythonGC.
        # It is very important to be a sorted list, so first elements to arrive are cleaned before,
        # n any deserialization from DB or parameter, objects deserialized first are referrers to
        # objects deserialized later. Second ones cannot be GC if first ones are not cleaned.
        # During GC,we should know that somehow. It's a hint but improves GC a lot.
        # Also, remember list must be thread-safe:
        # Lists themselves are thread-safe. In CPython the GIL protects against concurrent accesses to them
        self.retained_objects = list()
    
        # Store also the ObjectID, because it is the fastest way to check if a certain object
        # is there (we cannot rely on __eq__ operations on user-defined classes.
        self.retained_objects_id = set()
        
        """ Indicates if HeapManager is flushing all objects in Heap to disk. """
        self.is_flushing_all = False 
        
        """ Indicates if HeapManager is processing GC"""
        self.is_processing_gc = False
        self.exec_env = theruntime.get_execution_environment()
        self.logger.debug("EE HEAP MANAGER created for EE %s", self.exec_env.ee_name)

    def get_object_ids_retained(self):
        """
        @postcondition: get ids of objects retained in memory
        @return ids of objects retained in memory 
        """ 
        return self.inmemory_objects.keys()

    def add_to_heap(self, dc_object):
        """
        @postcondition: the object is added to dataClay's heap
        @param dc_object: object to add to the heap 
        """
        self._add_to_inmemory_map(dc_object)
        self.retain_in_heap(dc_object)
        # self.logger.debug("[==GC==] Added to heap object %s with address %s ", (dc_object.get_object_id(), str(id(dc_object))))

    def retain_in_heap(self, dc_object):
        """
        @postcondition: Add a new Hard reference to the object provided. All code in stubs/exec classes using objects in dataClayheap are
        using weak references. In order to avoid objects to be GC without a flush in DB, HeapManager has hard-references to
        them and is the only one able to release them. This function creates the hard-reference.
        @param dc_object: Object to retain. 
        """ 
        if dc_object.get_object_id() not in self.retained_objects_id:
            self.retained_objects_id.add(dc_object.get_object_id())
            self.retained_objects.append(dc_object)

    def release_from_heap(self, dc_obj):
        """ 
        @postcondition: Release hard reference to object provided. Without hard reference, the object can be Garbage collected
        @param dc_obj: object to release
        """
        self.logger.debug("[==GC==] Releasing object with id %s from retained map. ", dc_obj.get_object_id())
        try:
            self.retained_objects_id.remove(dc_obj.get_object_id())
            self.retained_objects.remove(dc_obj)        
        except Exception as e:
            self.logger.debug("[==GC==] ERROR Releasing object with id %s ", dc_obj.get_object_id())
    
    def __check_memory_pressure(self):
        """
        @postcondition: Check if memory is under pressure 
        @return TRUE if memory is under pressure. FALSE otherwise. 
        """
        """ 
        Memory management in Python involves a private heap containing all Python objects and data structures. 
        The management of this private heap is ensured internally by the Python memory manager. 
        The Python memory manager has different components which deal with various dynamic storage management aspects, 
        like sharing, segmentation, preallocation or caching.
        """
        virtual_mem = psutil.virtual_memory()
        self.logger.trace("[==GC==] Memory: %s", virtual_mem)
        return float(virtual_mem.percent) > (Configuration.MEMMGMT_PRESSURE_FRACTION * 100)

    def __check_memory_ease(self):
        """
        @postcondition: Check if memory is at ease
        @return TRUE if memory is at ease. FALSE otherwise.
        """

        # See __check_memory_pressure, as this is quite similar

        virtual_mem = psutil.virtual_memory()
        self.logger.trace("[==GC==] Memory: %s", virtual_mem)
        return float(virtual_mem.percent) < (Configuration.MEMMGMT_EASE_FRACTION * 100)

    def __nullify_object(self, dc_object):
        """
        @postcondition: Set all fields to none to allow GC action 
        """ 
        
        metaclass = dc_object.get_class_extradata()
        self.logger.debug("[==GC==] Going to clean object %r", dc_object)

        # Put here because it is critical path and I prefer to have a single isEnabledFor
        # instead of checking it for each element
        if self.logger.isEnabledFor(logging.DEBUG):
            held_objects = WeakValueDictionary()

            o = None
            prop_name_list = metaclass.properties.keys()

            self.logger.debug("The following attributes will be nullified from object %r: %s",
                              dc_object, ", ".join(prop_name_list))

            for prop_name in prop_name_list:
                real_prop_name = "%s%s" % (DCLAY_PROPERTY_PREFIX, prop_name)

                try:
                    o = object.__getattribute__(dc_object, real_prop_name)
                    held_objects[prop_name] = o
                except TypeError:
                    # Some objects cannot be weakreferenced, but we can typically ignore them
                    self.logger.trace("Ignoring attribute %s of type %s",
                                      prop_name, type(o))

            # Ensure we don't keep that as a dangling active backref
            del o

        # Critical path, keep it short!
        for prop_name in metaclass.properties.keys():
            real_prop_name = "%s%s" % (DCLAY_PROPERTY_PREFIX, prop_name)
            object.__setattr__(dc_object, real_prop_name, None)

        if self.logger.isEnabledFor(logging.DEBUG):
            # held_objects variable will be defined when TRACE-enabled.
            held_attr_names = held_objects.keys()

            if held_attr_names:
                self.logger.debug("The following attributes of object %r still have a backref active: %s",
                                  dc_object, ", ".join(held_attr_names))
            else:
                self.logger.debug("The garbage collector seems to have cleaned all the nullified attributes on %r",
                                  dc_object)

    def __clean_object(self, dc_object):
        """
        @postcondition: Clean object (except if not loaded or being used). Cleaning means set all fields to None to allow
        GC to work.
        @param dc_object: Object to clean.
        """
        
        """
        Lock object (not locking executions!)
        Lock is needed in case object is being nullified and some threads requires to load it from disk.
        """
        object_id = dc_object.get_object_id()
        self.runtime.lock(object_id)
        try:
            
            is_loaded = dc_object.is_loaded()
            if not is_loaded: 
                self.logger.trace("[==GC==] Not collecting since not loaded.")
                return
            
            """ Set loaded flag to false, any current execution that wants to get/set a field must try to load
            object from DB, and lock will control that object is not being cleaned """
            self.logger.debug("[==GC==] Setting loaded to false from gc %s" % str(object_id))

            dc_object.set_loaded(False) 
            
            if dc_object.is_dirty() or dc_object.is_pending_to_register(): 
                # Update it 
                self.logger.debug("[==GC==] Updating object %s ", dc_object.get_object_id())
                self.gc_collect_internal(dc_object)
            
            self.logger.debug("[==GC==] Cleaning object %s", dc_object.get_object_id())
            
            self.__nullify_object(dc_object)
            
            """ Object is not dirty anymore """
            dc_object.set_dirty(False)
            
            """
            VERY IMPORTANT (RACE CONDITION)
            If some object was cleaned and removed from GC retained refs, it does NOT mean it was removed
            from Weak references Heap because we will ONLY remove an entry in that Heap if the GC removed it.
            So, if some execution is requested after we remove an entry from retained refs (we cleaned and send
            the object to disk), we check if the
            object is in Heap (see executeImplementation as an example) and therefore, we created a new reference
            making impossible for GC to clean the reference. We will add the object to retained refs
            again once it is deserialized from DB. See DeserializationLib. It's the best solution without Lockers 
            in get and remove in Heap.

            Remove it from Retained refs to allow GC action.
            """
            self.release_from_heap(dc_object)

        finally:
            self.runtime.unlock(object_id)

    def gc_collect_internal(self, object_to_update):
        """
        @postcondition: Update object in db or store it if volatile (and register in LM)
        @param object_to_update: object to update
        """
        try:
            self.logger.debug("[==GCUpdate==] Updating object %s", object_to_update.get_object_id())
            """ Call EE update """ 
            if object_to_update.is_pending_to_register():
                obj_bytes = SerializationLibUtilsSingleton.serialize_for_db_gc(object_to_update, False, None)
                self.logger.debug("[==GCUpdate==] Pending to register in LM ")
                self.exec_env.register_and_store_pending(object_to_update, obj_bytes, True)
            elif object_to_update.is_dirty():
                obj_bytes = SerializationLibUtilsSingleton.serialize_for_db_gc(object_to_update, False, None)
                self.logger.debug("[==GCUpdate==] Updated dirty object %s ", object_to_update.get_object_id())
                self.runtime.update_to_sl(object_to_update.get_object_id(), obj_bytes, True)
            else: 
                # TODO: how to check if GlobalGC is enabled? 
                obj_bytes = SerializationLibUtilsSingleton.serialize_for_db_gc_not_dirty(object_to_update, False, None)
                if obj_bytes is not None:
                    ref_counting_bytes = DeserializationLibUtilsSingleton.extract_reference_counting(obj_bytes)
                    self.runtime.update_to_sl(object_to_update.get_object_id(), ref_counting_bytes, False)

        except: 
            # do nothing
            traceback.print_exc()
        """ TODO: set datasetid for GC if set by user """

    def run_task(self): 
        """
        @postcondition: Check Python VM's memory pressure and clean if necessary. Cleaning means flushing objects, setting 
        all fields to none (to allow GC to work better) and remove from retained references. If volatile or pending to register,
        we remove it once registered.
        """
        if self.is_flushing_all or self.is_processing_gc:
            self.logger.debug("[==GC==] Not running since is being processed or flush all is being done")
            return 
        # No race condition possible here since there is a time interval for run_gc that MUST be > than time spend to check 
        # flag and set to True. Should we do it atomically? 
        self.is_processing_gc = True
        self.logger.trace("[==GC==] Running GC")

        self.exec_env.prepareThread()
        is_pressure = self.__check_memory_pressure()
        if is_pressure:
            self.logger.verbose("System memory is under pressure, proceeding to clean up objects")
            if self.logger.isEnabledFor(logging.DEBUG) and tracemalloc is not None and tracemalloc.is_tracing():
                self.logger.debug("Doing a snapshot...")
                snapshot = tracemalloc.take_snapshot()
                top_stats = snapshot.statistics('lineno')

                print("[ Top 10 ]")
                for stat in top_stats[:10]:
                    print(stat)

            """
            TODO: CORRECT THESE LOGS
            cur_frame = sys._getframe(0)
            self.logger.debug("[==GC==] Is enabled? %s", str(gc.isenabled()))
            self.logger.debug("[==GC==] Is retained objects tracked? %s", str(gc.is_tracked(self.retained_objects)))
            self.logger.debug("[==GC==] Inmemory map %s", str(id(HeapManager.get_heap(self))))
            self.logger.debug("[==GC==] Retained map %s", str(id(self.retained_objects)))
            self.logger.debug("[==GC==] Threshold: %s", str(gc.get_threshold()))
            self.logger.debug("[==GC==] Count: %s", str(gc.get_count()))
            self.logger.debug("[==GC==] Current frame ID: %s", str(id(cur_frame)))
            """

            # Copy the references in order to process it in a plain fashion
            retained_objects_copy = self.retained_objects[:]
            self.logger.debug("Starting iteration with #%d retained objects", len(self.retained_objects))
            while retained_objects_copy:
                # We iterate through while-pop to ensure that the reference is freed
                dc_obj = retained_objects_copy.pop()

                if dc_obj.get_memory_pinned():
                    self.logger.trace("Object %r is memory pinned, ignoring it", dc_obj)
                    continue
                
                """ 
                self.logger.debug("[==GC==] Object address in memory: %s", str(id(dc_obj)))
                self.logger.debug("[==GC==] Is tracked? %s", str(gc.is_tracked(dc_obj)))
                for r in gc.get_referents(dc_obj):
                    if r == self.retained_objects:
                        self.logger.debug("[==GC==] REFERENT BEFORE CLEAN FOR %s is retained map. ", (dc_obj.get_object_id()))
                    elif r == HeapManager.get_heap(self):
                        self.logger.debug("[==GC==] REFERENT BEFORE CLEAN FOR %s is inmemory map. ", (dc_obj.get_object_id())) 
                    else:
                        self.logger.debug("[==GC==] REFERENT BEFORE CLEAN FOR %s is: %s ", (dc_obj.get_object_id(), str(id(r))))
                        # self.logger.debug("[==GC==] REFERENT BEFORE CLEAN FOR %s is: %s ", (dc_obj.get_object_id(), pprint.pformat(r)))
                for r in gc.get_referrers(dc_obj):
                    if r == self.retained_objects:
                        self.logger.debug("[==GC==] REFERRER BEFORE CLEAN FOR %s is retained map. ", (dc_obj.get_object_id()))
                    elif r == HeapManager.get_heap(self):
                        self.logger.debug("[==GC==] REFERRER BEFORE CLEAN FOR %s is inmemory map. ", (dc_obj.get_object_id()))
                    else:
                        self.logger.debug("[==GC==] REFERRER BEFORE CLEAN FOR %s is: %s ", (dc_obj.get_object_id(), str(id(r))))
                        # self.logger.debug("[==GC==] REFERRER BEFORE CLEAN FOR %s is: %s ", (dc_obj.get_object_id(), pprint.pformat(r)))
                """
                self.__clean_object(dc_obj)
                """
                for r in gc.get_referents(dc_obj):
                    if r == self.retained_objects:
                        self.logger.debug("[==GC==] REFERENT FOR %s is retained map. ", (dc_obj.get_object_id()))
                    elif r == HeapManager.get_heap(self):
                        self.logger.debug("[==GC==] REFERENT FOR %s is inmemory map. ", (dc_obj.get_object_id()))
                    else:
                        self.logger.debug("[==GC==] REFERENT FOR %s is: %s ", (dc_obj.get_object_id(), str(id(r))))
                        # self.logger.debug("[==GC==] REFERENTS FOR %s is: %s ", (dc_obj.get_object_id(), pprint.pformat(r)))
                """
                """
                for r in gc.get_referrers(dc_obj):
                    if r == self.retained_objects:
                        self.logger.debug("[==GC==] REFERRER FOR %s is retained map. ", (dc_obj.get_object_id()))
                    elif r == HeapManager.get_heap(self):
                        self.logger.debug("[==GC==] REFERRER FOR %s is inmemory map. ", (dc_obj.get_object_id()))
                    else:
                        self.logger.debug("[==GC==] ID REFERRER FOR %s is: %s ", (dc_obj.get_object_id(), str(id(r))))
                        self.logger.debug("[==GC==] REFERRER FOR %s are: %s ", (dc_obj.get_object_id(), pprint.pformat(r)))
                """
                del dc_obj  # Remove reference from Frame

                # Big heaps
                n = gc.collect()
                if self.logger.isEnabledFor(logging.DEBUG):
                    if n > 0:
                        self.logger.debug("[==GC==] Collected %d", n)
                    else:
                        self.logger.trace("[==GC==] No objects collected")
                    if gc.garbage:
                        self.logger.debug("[==GC==] Uncollectable: %s", gc.garbage)
                    else:
                        self.logger.trace("[==GC==] No uncollectable objects")
            
                # Check memory 
                at_ease = self.__check_memory_ease()
                if at_ease:
                    self.logger.trace("[==GC==] Not collecting since memory is 'at ease' now")
                    break
                if self.is_flushing_all:
                    self.logger.debug("[==GC==] Interrupted due to flush all.")
                    break
            else:
                # This block is only entered when the while has finished (with no break statement)
                # which means that all the objects have been iterated
                self.logger.warning("I did my best and ended up cleaning all retained_objects. This typically means"
                                    " that there is a huge global memory pressure. Problematic.")
            self.logger.debug("Finishing iteration with #%d retained objects", len(self.retained_objects))

            self.cleanReferencesAndLockers()

            # For cyclic references
            n = gc.collect()

            if self.logger.isEnabledFor(logging.DEBUG):
                if retained_objects_copy:
                    self.logger.debug("There are #%d remaining objects after Garbage Collection",
                                      len(retained_objects_copy))

                if n > 0:
                    self.logger.debug("[==GC==] Finally Collected %d", n)
                else:
                    self.logger.trace("[==GC==] No objects collected")
                if gc.garbage:
                    self.logger.debug("[==GC==] Uncollectable: %s", gc.garbage)
                else:
                    self.logger.trace("[==GC==] No uncollectable objects")

            del retained_objects_copy
            
            if self.logger.isEnabledFor(logging.DEBUG) and tracemalloc is not None and tracemalloc.is_tracing():
                self.logger.debug("Doing a snapshot...")
                snapshot2 = tracemalloc.take_snapshot()

                top_stats = snapshot2.compare_to(snapshot, 'lineno')

                print("[ Top 10 differences ]")
                for stat in top_stats[:10]:
                    print(stat)

        self.is_processing_gc = False
        return
        
    def flush_all(self):
        """
        @postcondition: Stores all objects in memory into disk. This function is usually called at shutdown of the 
        execution environment. 
        """
        # If there is another flush, return
        if self.is_flushing_all: 
            return 
        
        self.is_flushing_all = True
        
        # If there is GC being processed, wait
        while self.is_processing_gc: 
            time.sleep(self.TIME_WAIT_FOR_GC_TO_FINISH)

        self.logger.debug("[==FlushAll==] Number of objects in Heap: %s", self.heap_size())
        for object_to_update in self.retained_objects:
            self.gc_collect_internal(object_to_update)
                
        return



"""Internal module used by both client and servers.

The classes and functions in the dataClay module are available (when this
makes sense) to both the dataClay client and the Python dataClay Execution
Environment.

The "client" version is available at `dataclay` package, which works as en entry
point for all the commonruntime user-friendly functions.
"""
from contextlib import contextmanager
from distutils.util import strtobool
import logging
import logging.config
import os
import yaml

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

# Manually forcing the dataclay root logger here
logger = logging.getLogger("dataclay")
LOGGING_FORMAT = '%(asctime)s [%(processName)s] [%(threadName)s] [%(name)s] [%(levelname)s] [%(module)s:%(lineno)d] %(message)s'

###################################################################
# We like to have a little bit more of finesse wrt debug levels

# for lower-than-debug messages
TRACE_LEVEL_NUM = 5
logging.addLevelName(TRACE_LEVEL_NUM, "TRACE")

# for higher-than-debug but not printed by default
VERB_LEVEL_NUM = 15
logging.addLevelName(VERB_LEVEL_NUM, "VERBOSE")


def trace(self, message, *args, **kws):
    if self.isEnabledFor(TRACE_LEVEL_NUM):
        self._log(TRACE_LEVEL_NUM, message, args, **kws)


def verbose(self, message, *args, **kws):
    if self.isEnabledFor(VERB_LEVEL_NUM):
        self._log(VERB_LEVEL_NUM, message, args, **kws)


# And monkey patch the logging library also
logging.TRACE = TRACE_LEVEL_NUM
logging.VERB = VERB_LEVEL_NUM
logging.VERBOSE = VERB_LEVEL_NUM
logging.Logger.verb = verbose
logging.Logger.verbose = verbose
logging.Logger.trace = trace


def _get_logging_dict_config():
    """Return the dictionary config for the logging.

    This function provides either a sensible default (generated here) or loads
    a YAML file provided through the environment variable called
    "DATACLAY_LOGGING_CONFIG".
    
    :return: A dictionary that can be used with logging.dictConfig
    """
    file_config = os.getenv('DATACLAY_LOGGING_CONFIG')
    if file_config:
        dict_config = yaml.load(open(file_config, 'r'))

        handlers = dict_config.get("handlers", tuple())
        for h in handlers:
            try:
                template = handlers[h].pop("filename_template")
            except KeyError:
                pass
            else:
                new_filename = template.format(PID=os.getpid(), **os.environ)
                logger.debug("Handler %s is using `%s`, formatter from `%s`",
                             h, new_filename, template)
                # Ensure that folder exists
                try:
                    os.makedirs(os.path.dirname(new_filename))
                except OSError:
                    # If path exists, ignore
                    pass
                handlers[h]["filename"] = new_filename
        return dict_config
    else:
        debug = strtobool(os.getenv('DEBUG', "False"))
        level_name = os.getenv('LOGLEVEL')

        # Note that the priority between DEBUG and LOGLEVEL is not random, albeit arbitrary
        if level_name == "VERBOSE":
            level = logging.VERBOSE
        elif level_name == "DEBUG":
            level = logging.DEBUG
        elif level_name == "TRACE":
            level = logging.TRACE
        elif debug:
            level = logging.DEBUG
        elif level_name == "INFO":
            level = logging.INFO
        else:
            level = logging.WARNING

        return {
            "version": 1,
            "formatters": {
                "simple": {
                    "format": LOGGING_FORMAT,
                },
            },
            "handlers": {
                "dclay_console": {
                    "class": "logging.StreamHandler",
                    "level": level,
                    "formatter": "simple",
                    "stream": "ext://sys.stdout",
                }
            },
            "loggers": {
                # TODO: Add grpc logs
                "dataclay": {
                    "level": level,
                    "handlers": [
                        "dclay_console",
                    ]
                }
            }
        }


def initialize():
    """Initialize the dataClay frame (logging, tracing and constants).

    The caller should, prior to this initialize, set the ConfigOptions to valid
    values. After this initialize, the dataClay library is ready to go.
    """
    config_kwargs = _get_logging_dict_config()
    config_kwargs["disable_existing_loggers"] = False
    logging.config.dictConfig(config_kwargs)

    logger.verbose("Starting dataClay library")
    logger.verbose("Debug output seems to be enabled")

        
@contextmanager
def size_tracking(io_file):
    """Track the bytes written into a certain seekable I/O file.
    :param io_file: The I/O file being written inside the with statement.
    """
    # Hack a little bit a circular import
    from dataclay.serialization.python.lang.IntegerWrapper import IntegerWrapper

    start_track = io_file.tell()
    IntegerWrapper(32).write(io_file, 0)
    start_data = io_file.tell()
    yield
    end_data = io_file.tell()
    io_file.seek(start_track)
    IntegerWrapper(32).write(io_file, end_data - start_data)
    io_file.seek(end_data)




""" Class description goes here. """

from dataclay import DataClayObject
from importlib import import_module
import logging

from dataclay.commonruntime.ExecutionGateway import loaded_classes
from dataclay.util.management.classmgr.UserType import UserType
from dataclay.exceptions.exceptions import DataClayException
import traceback
__author__ = 'Alex Barcelo <alex.barcelo@bsc.es'
__copyright__ = '2016 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)


class MetaClassFactory(object):
    """Tracker of classes and generator of dataClay MetaClasses.

    One of the functions of this class is managing a set of classes, which
    may have cross dependencies between them. Additionally, this factory helps
    to manage a set of MetaClasses (prior to a registration process).

    The keyword parameters are used to complete the containers for the
    MetaClass being registered.
    """

    def __init__(self, namespace, responsible_account):
        """Simple class initialization.

        :param str namespace: The string of the namespace.
        :param str responsible_account: The registrator account (username).
        """
        self.classes = list()
        self.types = dict()

        self._responsible_account = responsible_account
        self._namespace = namespace

        self._prefix = None
        self._ignored_prefixes = -1

    def import_and_add(self, import_str):
        """Perform a import operation while adding classes.

        This method calls to importlib.import_module, while watching the
        StorageObject classes that are being loaded. All the classes that are
        loaded as a result of the import will be added to this factory.

        :param import_str: A string that can be used as parameter to import_module.
        """
        loaded_classes.clear()

        try:
            import_module(import_str)
        except ImportError as e:
            traceback.print_exc()
            logger.warning("Tried to import `%s` and failed, ignoring", import_str)
            logger.warning("Error: %s", e)
        else:
            for k in loaded_classes:
                if k.__module__.startswith("dataclay"):
                    # dataClay contrib classes should not be registered here
                    continue
                else:
                    self.add_class(k)

    def add_class(self, klass):
        """Add a class to this factory, from the class' Python object.

        Note that the caller provides de class, which should be an instance of
        ExecutionGateway.

        :param klass: The class object.
        """
        if not issubclass(klass, DataClayObject):
            raise DataClayException("Can only use DataClayObject classes")

        logger.verbose("Adding class %s to the MetaClassFactory", klass)
        class_container = klass._prepare_metaclass(
            self._namespace, self._responsible_account)

        # Save to the list, and bookmark the MetaClass
        # (for valid recursive behaviour, e.g. cycles)
        complete_name = class_container.name
        logger.debug("[add_class] Using `%s` as `name` field of Type",
                     complete_name)
        if complete_name not in self.types:
            self.types[complete_name] = UserType(
                signature="L{};".format(complete_name).replace(".", "/"),
                includes=[],
                namespace=self._namespace,
                typeName=complete_name)
        self.classes.append(class_container)

        parent = klass.__bases__[0]
        if parent is not DataClayObject:
            self.add_class(parent)

        logger.debug("Class %s finished", class_container.name)

    def __str__(self):
        return "MetaClass Factory containing:\n{}".format(str(self.classes))




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class Interface(ManagementObject):
    _fields = ["dataClayID",
               "providerAccountName",
               "namespace",
               "classNamespace",
               "className",
               "propertiesInIface",
               "operationsSignatureInIface",
               ]

    _internal_fields = ["providerAccountID",
                        "namespaceID",
                        "classNamespaceID",
                        "metaClassID",
                        "operationsIDs",
                        "propertiesIDs",
                        ]




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class Namespace(ManagementObject):
    _fields = ["dataClayID",
               "providerAccountName",
               "name",
               ]

    _internal_fields = ["responsible",
                        "importedInterfaces",
                        "language",
                        ]




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class MetaDataInfo(ManagementObject):
    _fields = ["dataClayID",
               "isReadOnly",
               "datasetID",
               "metaclassID",
               "locations",
               "aliases",
               "ownerID"
               ]

    _internal_fields = []




""" Class description goes here. """

"""Entry point for standalone dataClay Execution Environment server.

The main can be called easily through a

    python -m dclay_server
"""

import logging

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)


# We create a specific function that can be also run from importing the module (testing)
def run_main():
    # Current execution environment since they are initialized using environment variables and cannot be concurrently started in same host.
    from dataclay import initialize
    initialize()
    from dataclay.executionenv.server.ExecutionEnvironmentSrv import ExecutionEnvironmentSrv
    exec_env_srv = ExecutionEnvironmentSrv()
    exec_env_srv.start()

    
if __name__ == "__main__":
    run_main()
    



"""Management of Python Classes.

This module is responsible of management of the Class Objects. A central Python
Metaclass is responsible of Class (not object) instantiation.

Note that this managers also includes most serialization/deserialization code
related to classes and function call parameters.
"""
import inspect
import traceback
import logging
import copy
import re
from operator import attrgetter
from uuid import UUID

import six

if six.PY2:
    from cPickle import Pickler, Unpickler
elif six.PY3:
    from _pickle import Pickler, Unpickler
import uuid
from dataclay.DataClayObjMethods import dclayMethod
from dataclay.DataClayObjProperties import DCLAY_PROPERTY_PREFIX, PreprocessedProperty, DynamicProperty, \
    ReplicatedDynamicProperty
from dataclay.DataClayObjectExtraData import DataClayInstanceExtraData, DataClayClassExtraData
from dataclay.commonruntime.ExecutionGateway import ExecutionGateway, class_extradata_cache_client, \
    class_extradata_cache_exec_env
from dataclay.commonruntime.Runtime import getRuntime
from dataclay.commonruntime.RuntimeType import RuntimeType
from dataclay.serialization.lib.DeserializationLibUtils import DeserializationLibUtilsSingleton, PersistentLoadPicklerHelper
from dataclay.serialization.lib.SerializationLibUtils import SerializationLibUtilsSingleton, PersistentIdPicklerHelper
from dataclay.serialization.python.lang.BooleanWrapper import BooleanWrapper
from dataclay.serialization.python.lang.StringWrapper import StringWrapper
from dataclay.serialization.python.lang.IntegerWrapper import IntegerWrapper
from dataclay.serialization.python.lang.DCIDWrapper import DCIDWrapper
from dataclay.util.StubUtils import load_babel_data
from dataclay.util.classloaders.ClassLoader import load_metaclass
from dataclay.util.management.classmgr.Type import Type
from dataclay.util.management.classmgr.UserType import UserType
from dataclay.exceptions.exceptions import DataClayException, ImproperlyConfigured
import six

# Publicly show the dataClay method decorators
__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2016 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)

# For efficiency purposes compile the folowing regular expressions:
# (they return a tuple of two elements)
re_property = re.compile(r"(?:^\s*@dclayReplication\s*\(\s*(before|after)Update\s*=\s*'([^']+)'(?:,\s(before|after)Update='([^']+)')?(?:,\sinMaster='(True|False)')?\s*\)\n)?^\s*@ClassField\s+([\w.]+)[ \t]+([\w.\[\]<> ,]+)", re.MULTILINE)
re_import = re.compile(r"^\s*@d[cC]layImport(?P<from_mode>From)?\s+(?P<import>.+)$", re.MULTILINE)


def _get_object_by_id_helper(object_id, class_id, hint):
    """Helper method which can be pickled and used by DataClayObject.__reduce__"""
    return getRuntime().get_object_by_id(object_id, class_id, hint)


@six.add_metaclass(ExecutionGateway)
class DataClayObject(object):
    """Main class for Persistent Objects.

    Objects that has to be made persistent should derive this class (either
    directly, through the StorageObject alias, or through a derived class).
    """

    # Extradata of the object. Private field.
    __dclay_instance_extradata = None
    
    def initialize_object(self, new_object_id=None):
        """ 
        @postcondition: Initialize the object 
        @param new_object_id: Object Id of the object
        """
        if new_object_id is not None:  # TODO: remove this if once ExecutionGateway is not initializing object id twice
            self.set_object_id(new_object_id)
        getRuntime().add_to_heap(self)
        
    def initialize_object_as_persistent(self):
        """
        @postcondition: object is initialized as a persistent object.
        Flags for "persistent" state might be different in EE and client.
        """
        # TODO: improve this using an specialization (dgasull) 
        if getRuntime().is_exec_env(): 
            # *** Execution Environment flags
            self.set_persistent(True)

            # by default, loaded = true for volatiles created inside executions
            # this function (initialize as persistent) is used for objects being
            # deserialized and therefore they might be unloaded
            # same happens for pending to register flag.
            self.set_loaded(False)
            self.set_pending_to_register(False)
            
        else: 
            # *** Client flags
            self.set_persistent(True)

    def initialize_object_as_volatile(self):
        """
        @postcondition: Initialize object with state 'volatile' with proper flags. Usually, volatile state is created by a stub, app, exec
        class,.. See same function in DataClayExecutionObject for a different initialization. This design is intended to be
        clear with object state.
        """
        # TODO: improve this using an specialization (dgasull) 
        if getRuntime().is_exec_env(): 
            # *** Execution Environment flags
            self.set_persistent(True)  # All objects in the EE are persistent
            self.set_loaded(True)
            self.set_pending_to_register(True)
            self.set_hint(getRuntime().get_hint())
            self.set_owner_session_id(getRuntime().get_session_id())
    
    @classmethod
    def get_class_extradata(cls):
        classname = cls.__name__
        module_name = cls.__module__
        full_name = "%s.%s" % (module_name, classname)

        # Check if class extradata is in cache.
        if getRuntime().current_type == RuntimeType.client:
            dc_ced = class_extradata_cache_client.get(full_name)
        else:
            dc_ced = class_extradata_cache_exec_env.get(full_name)

        if dc_ced is not None:
            return dc_ced

        logger.verbose('Proceeding to prepare the class `%s` from the ExecutionGateway',
                       full_name)
        logger.debug("The RuntimeType is: %s",
                     "client" if getRuntime().current_type == RuntimeType.client else "not client")

        dc_ced = DataClayClassExtraData(
            full_name=full_name,
            classname=classname,
            namespace=module_name.split('.', 1)[0],
            properties=dict(),
            imports=list(),
        )

        if getRuntime().current_type == RuntimeType.client:
            class_stubinfo = None

            try:
                all_classes = load_babel_data()

                for c in all_classes:
                    if "%s.%s" % (c.namespace, c.className) == full_name:
                        class_stubinfo = c
                        break
            except ImproperlyConfigured:
                pass

            if class_stubinfo is None:
                # Either ImproperlyConfigured (not initialized) or class not found:
                # assuming non-registered class

                # Let's navigate all the heritance, and aggregate all the information that can be found.
                # FIXME: This code is NOT considering the scenario in which a class is extending an already registered one.
                # FIXME: Behaviour when an attribute is defined in more than one class is not tested nor defined right now.
                for current_cls in inspect.getmro(cls):
                    # Ignore `object` and also ignore DataClayObject itself
                    if current_cls in [object, DataClayObject]:
                        continue

                    # Prepare properties from the docstring
                    doc = current_cls.__doc__  # If no documentation, consider an empty string
                    if doc is None:
                        doc = ""
                    property_pos = 0

                    for m in re_property.finditer(doc):
                        # declaration in the form [ 'before|after', 'method', 'before|after', 'method', 'name', 'type' ]
                        declaration = m.groups()
                        prop_name = declaration[-2]
                        prop_type = declaration[-1]

                        beforeUpdate = declaration[1] if declaration[0] == 'before' \
                            else declaration[3] if declaration[2] == 'before' else None

                        afterUpdate = declaration[1] if declaration[0] == 'after' \
                            else declaration[3] if declaration[2] == 'after' else None

                        inMaster = declaration[4] == 'True'

                        current_type = Type.build_from_docstring(prop_type)

                        logger.trace("Property `%s` (with type signature `%s`) ready to go",
                                    prop_name, current_type.signature)

                        dc_ced.properties[prop_name] = PreprocessedProperty(
                            name=prop_name,
                            position=property_pos,
                            type=current_type,
                            beforeUpdate=beforeUpdate,
                            afterUpdate=afterUpdate,
                            inMaster=inMaster)

                        # Keep the position tracking (required for other languages compatibility)
                        property_pos += 1

                        # Prepare the `property` magic --this one without getter and setter ids
                        # dct[prop_name] = DynamicProperty(prop_name)
                        setattr(cls, prop_name, DynamicProperty(prop_name))
                        # WARNING: This is done in `cls` and not in `current_cls` deliberately.

                    for m in re_import.finditer(doc):
                        gd = m.groupdict()

                        if gd["from_mode"]:
                            import_str = "from %s\n" % gd["import"]
                        else:
                            import_str = "import %s\n" % gd["import"]

                        dc_ced.imports.append(import_str)

            else:
                logger.debug("Loading a class with babel_data information")

                dc_ced.class_id = class_stubinfo.classID
                dc_ced.stub_info = class_stubinfo

                # WIP WORK IN PROGRESS (because all that is for the ancient StubInfo, not the new one)

                # Prepare the `property` magic --in addition to prepare the properties dictionary too
                for i, prop_name in enumerate(dc_ced.stub_info.propertyListWithNulls):

                    if prop_name is None:
                        continue

                    prop_info = class_stubinfo.properties[prop_name]
                    if prop_info.beforeUpdate is not None or prop_info.afterUpdate is not None:
                        setattr(cls, prop_name,
                                ReplicatedDynamicProperty(
                                    prop_name,
                                    prop_info.beforeUpdate,
                                    prop_info.afterUpdate,
                                    prop_info.inMaster
                                ))

                    else:
                        # dct[prop_name] = DynamicProperty(prop_name)
                        setattr(cls, prop_name, DynamicProperty(prop_name))

                    dc_ced.properties[prop_name] = PreprocessedProperty(name=prop_name, position=i,
                                                                        type=prop_info.propertyType,
                                                                        beforeUpdate=prop_info.beforeUpdate,
                                                                        afterUpdate=prop_info.afterUpdate,
                                                                        inMaster=prop_info.inMaster)

        elif getRuntime().current_type == RuntimeType.exe_env:
            logger.verbose("Seems that we are a DataService, proceeding to load class %s",
                           dc_ced.full_name)
            namespace_in_classname, dclay_classname = dc_ced.full_name.split(".", 1)
            if namespace_in_classname != dc_ced.namespace:
                raise DataClayException("Namespace in ClassName: %s is different from one in ClassExtraData: %s",
                                        namespace_in_classname, dc_ced.namespace)
            mc = load_metaclass(dc_ced.namespace, dclay_classname)
            dc_ced.metaclass_container = mc
            dc_ced.class_id = mc.dataClayID

            # Prepare the `property` magic --in addition to prepare the properties dictionary too
            for prop_info in dc_ced.metaclass_container.properties:
                if prop_info.beforeUpdate is not None or prop_info.afterUpdate is not None:
                    setattr(cls, prop_info.name,
                            ReplicatedDynamicProperty(
                                prop_info.name,
                                prop_info.beforeUpdate,
                                prop_info.afterUpdate,
                                prop_info.inMaster
                            ))
                else:
                    setattr(cls, prop_info.name, DynamicProperty(prop_info.name))

                dc_ced.properties[prop_info.name] = PreprocessedProperty(
                    name=prop_info.name,
                    position=prop_info.position,
                    type=prop_info.type,
                    beforeUpdate=prop_info.beforeUpdate,
                    afterUpdate=prop_info.afterUpdate,
                    inMaster=prop_info.inMaster)
        else:
            raise RuntimeError("Could not recognize RuntimeType %s", getRuntime().current_type)

        # Update class extradata cache.
        if getRuntime().current_type == RuntimeType.client:
            class_extradata_cache_client[full_name] = dc_ced
        else:
            class_extradata_cache_exec_env[full_name] = dc_ced

        return dc_ced

    def _populate_internal_fields(self, deserializing=False, **kwargs):
        logger.debug("Populating internal fields for the class. Provided kwargs: %s deserializing=%s", kwargs, str(deserializing))

        # Mix default values with the provided ones through kwargs
        fields = {
            "persistent_flag": False,
            "object_id": uuid.uuid4(),
            "dataset_id": None,
            "loaded_flag": True,
            "pending_to_register_flag": False,
            "dirty_flag": False,
            "memory_pinned": False,
        }
        fields.update(kwargs)

        # Store some extradata in the class
        instance_dict = object.__getattribute__(self, "__dict__")
        instance_dict["_DataClayObject__dclay_instance_extradata"] = DataClayInstanceExtraData(**fields)

        """
        TODO: get_class_extradata function is adding DynamicProperties to class (not to instance!) so it is needed 
        to be called. Please, use a better function for that. 
        """
        instance_dict["_dclay_class_extradata"] = self.get_class_extradata()
        
        # Initialize object
        self.initialize_object()
        if not deserializing: 
            """ object created during executions is volatile. """
            self.initialize_object_as_volatile()

    def get_location(self):
        """Return a single (random) location of this object."""
        return getRuntime().get_location(self.__dclay_instance_extradata.object_id)

    def get_master_location(self):
        """Return the uuid relative to the master location of this object."""
        return self.__dclay_instance_extradata.master_location

    def set_master_location(self, eeid):
        """Set the master location of this object."""
        if not isinstance(eeid, UUID):
            raise AttributeError("The master location should be the ExecutionEnvironmentID, "
                                 "instead we received: %s" % eeid)
        self.__dclay_instance_extradata.master_location = eeid
        
    def get_all_locations(self):
        """Return all the locations of this object."""
        return getRuntime().get_all_locations(self.__dclay_instance_extradata.object_id)
    
    def new_replica(self, backend_id=None, recursive=True):
        getRuntime().new_replica(self.get_object_id(), self.get_class_extradata().class_id,
                                 self.get_hint(), backend_id, recursive)

    def new_version(self, backend_id):
        return getRuntime().new_version(self.get_object_id(), self.get_class_extradata().class_id,
                                        self.get_hint(), backend_id)

    def consolidate_version(self, version_info):
        getRuntime().consolidate_version(version_info)

    def make_persistent(self, alias=None, backend_id=None, recursive=True):
        if alias is "":
            raise AttributeError('Alias cannot be empty')
        getRuntime().make_persistent(self, alias=alias, backend_id=backend_id, recursive=recursive)

    def get_execution_environments_info(self):
        return getRuntime().get_execution_environments_info()

    @classmethod
    def dc_clone_by_alias(cls, alias, recursive=False):
        o = cls.get_by_alias(alias)
        return o.dc_clone(recursive)

    def dc_clone(self, recursive=False):
        """
        @postcondition: Returns a non-persistent object as a copy of the current object
        @return: DataClayObject non-persistent instance
        """
        return getRuntime().get_copy_of_object(self, recursive)

    @classmethod
    def dc_update_by_alias(cls, alias, from_object):
        o = cls.get_by_alias(alias)
        return o.dc_update(from_object)

    def dc_update(self, from_object):
        """
        @postcondition: Updates all fields of this object with the values of the specified object
        @param from_object: instance from which values must be retrieved to set fields of current object
        """
        if from_object is None:
            return
        else:
            getRuntime().update_object(self, from_object)
            
    def dc_put(self, alias, backend_id=None, recursive=True):
        if alias is None or alias is "":
            raise AttributeError('Alias cannot be null or empty')
        getRuntime().make_persistent(self, alias=alias, backend_id=backend_id, recursive=recursive)
            
    def set_all(self, from_object):
        properties = sorted(self.get_class_extradata().properties.values(),
                key=attrgetter('position'))

        logger.verbose("Set all properties from object %s", from_object.get_object_id())

        for p in properties:
            value = getattr(from_object, p.name)
            setattr(self, p.name, value)

    def get_object_id(self):
        """
        @postcondition: Return object id of the object
        @return Object id
        """
        return self.__dclay_instance_extradata.object_id
    
    def set_object_id(self, new_object_id):
        """
        @postcondition: Set object id of the object
        @param new_object_id: object id
        """
        self.__dclay_instance_extradata.object_id = new_object_id

    def _update_object_id(self, new_object_id):
        """
        @postcondition: Set a new object id for the object and
                        calls an update of its heap references
        @param new_object_id: object id
        """
        if self.is_persistent():
            raise DataClayException("Cannot change the id of a persistent object")
        old_object_id = self.get_object_id()
        self.__dclay_instance_extradata.object_id = new_object_id
        getRuntime().update_object_id(old_object_id, new_object_id)

    def get_memory_pinned(self):
        """
        @postcondition: Return the memory pinned flag of the object
        @return Object id
        """
        return self.__dclay_instance_extradata.memory_pinned

    def set_memory_pinned(self, new_memory_pinned):
        """
        @postcondition: Set memory pinned flag of the object
        @param new_memory_pinned: memory pinned flag
        """
        self.__dclay_instance_extradata.memory_pinned = new_memory_pinned

    def get_dataset_id(self):
        """
        @postcondition: Return dataset id of the object 
        @return Data set id
        """
        return self.__dclay_instance_extradata.dataset_id
    
    def set_dataset_id(self, new_dataset_id):
        """
        @postcondition: Set dataset id of the object
        @param new_dataset_id: dataset id
        """
        self.__dclay_instance_extradata.dataset_id = new_dataset_id

    def getID(self):
        """Return the string representation of the persistent object for COMPSs.

        dataClay specific implementation: The objects are internally represented
        through ObjectID, which are UUID. In addition to that, some extra fields
        are added to the representation. Currently, a "COMPSs ID" will be:

            <objectID>:<backendID|empty>:<classID>

        In which all ID are UUID and the "hint" (backendID) can be empty.

        If the object is NOT persistent, then this method returns None.
        """
        if self.is_persistent():
            hint = self.__dclay_instance_extradata.execenv_id or ""

            return "%s:%s:%s" % (
                self.__dclay_instance_extradata.object_id,
                hint,
                self.get_class_extradata().class_id
            )
        else:
            return None

    @classmethod
    def get_object_by_id(cls, object_id, *args, **kwargs):
        return getRuntime().get_object_by_id(object_id, *args, **kwargs)

    @classmethod
    def get_by_alias(cls, alias, safe=True):
        return getRuntime().get_by_alias(alias, cls.get_class_extradata().class_id, safe)

    @classmethod
    def delete_alias(cls, alias):
        return getRuntime().delete_alias(alias)

    def set_persistent(self, ispersistent):
        """
        @postcondition: Set value of persistent flag in the object
        @param ispersistent: value to set
        """
        self.__dclay_instance_extradata.persistent_flag = ispersistent

    def set_loaded(self, isloaded):
        """
        @postcondition: Set value of loaded flag in the object
        @param isloaded: value to set
        """
        logger.verbose("Setting loaded to `%s` for object %s", isloaded, self.get_object_id())
        self.__dclay_instance_extradata.loaded_flag = isloaded

    def is_persistent(self):
        """
        @postcondition: Return TRUE if object is persistent. FALSE otherwise. 
        @return is persistent flag
        """
        return self.__dclay_instance_extradata.persistent_flag

    def is_loaded(self):
        """
        @postcondition: Return TRUE if object is loaded. FALSE otherwise. 
        @return is loaded flag
        """
        return self.__dclay_instance_extradata.loaded_flag
    
    def get_owner_session_id(self):
        """
        @postcondition: Get owner session id
        @return Owner session id
        """
        return self.__dclay_instance_extradata.owner_session_id
    
    def set_owner_session_id(self, the_owner_session_id):
        """
        @postcondition: Set ID of session of the owner of the object. 
        @param the_owner_session_id: owner session id 
        """ 
        self.__dclay_instance_extradata.owner_session_id = the_owner_session_id
    
    def is_pending_to_register(self):
        """
        @postcondition: Return TRUE if object is pending to register. FALSE otherwise. 
        @return is pending to register flag
        """
        return self.__dclay_instance_extradata.pending_to_register_flag
    
    def set_pending_to_register(self, pending_to_register):
        """
        @postcondition: Set pending to register flag. 
        @param pending_to_register: The flag to set
        """
        self.__dclay_instance_extradata.pending_to_register_flag = pending_to_register
    
    def is_dirty(self):
        """
        @postcondition: Return TRUE if object is dirty (it was modified). FALSE otherwise. 
        @return dirty flag
        """
        return self.__dclay_instance_extradata.dirty_flag
    
    def set_dirty(self, dirty_value):
        """
        @postcondition: Set dirty flag. 
        @param dirty_value: The flag to set
        """
        self.__dclay_instance_extradata.dirty_flag = dirty_value
    
    def get_hint(self):
        """
        @postcondition: Get hint
        @return Hint
        """
        return self.__dclay_instance_extradata.execenv_id
    
    def set_hint(self, new_hint):
        """
        @postcondition: Set hint
        @param new_hint: value to set
        """
        self.__dclay_instance_extradata.execenv_id = new_hint

    def federate(self, ext_dataclay_id, recursive=True):
        """
        @postcondition: Federates this object with an external dataClay instance
        @param ext_dataclay_id: id of the external dataClay instance
        @param recursive: Indicates if all sub-objects must be federated as well.
        """
        getRuntime().federate_object(self.get_object_id(), ext_dataclay_id, recursive,
                                     self.get_class_extradata().class_id, self.get_hint())
    
    def unfederate(self, ext_dataclay_id=None, recursive=True):
        """
        @postcondition: Unfederate this object with an external dataClay instance
        @param ext_dataclay_id: id of the external dataClay instance (none means to unfederate with all dcs)
        @param recursive: Indicates if all sub-objects must be unfederated as well.
        """
        if ext_dataclay_id is not None:
            getRuntime().unfederate_object(self.get_object_id(), ext_dataclay_id, recursive)
        else:
            getRuntime().unfederate_object_with_all_dcs(self.get_object_id(), recursive)          

    def get_external_dataclay_id(self, dcHost, dcPort):
        return getRuntime().ready_clients["@LM"].get_external_dataclay_id(dcHost, dcPort)

    def set_in_backend(self, backend_id, field_name, value):
        from dataclay.DataClayObjProperties import DCLAY_SETTER_PREFIX
        return getRuntime().run_remote(self.get_object_id(), backend_id, DCLAY_SETTER_PREFIX + field_name, value)

    def get_external_dataclay_info(self, dataclay_id):
        """ Get external dataClay information
        :param dataclay_id: external dataClay ID
        :return: DataClayInstance information
        :type dataclay_id: UUID
        :rtype: DataClayInstance
        """
        return getRuntime().get_external_dataclay_info(dataclay_id)

    def get_federation_source(self):
        """ Retrieve dataClay instance id where the object comes from or NULL
        :return: dataClay instance ids where this object is federated
        :rtype: UUID
        """
        return getRuntime().get_external_source_of_dataclay_object(self.get_object_id())

    def get_federation_targets(self):
        """ Retrieve dataClay instances ids where the object is federated
        :return: dataClay instances ids where this object is federated
        :rtype: set of UUID
        """
        return getRuntime().get_dataclays_object_is_federated_with(self.get_object_id())

    def set_in_dataclay_instance(self, dc_info, field_name, params):
        from dataclay.DataClayObjProperties import DCLAY_SETTER_PREFIX
        getRuntime().synchronize_federated(self, params, DCLAY_SETTER_PREFIX + field_name, dc_info)

    def serialize(self, io_file, ignore_user_types, iface_bitmaps,
                  cur_serialized_objs, pending_objs, reference_counting):
        # Reference counting information
        # First integer represent the position in the buffer in which
        # reference counting starts. This is done to avoid "holding"
        # unnecessary information during a store or update in disk.

        # in new serialization, this will be done through padding
        # TODO: use padding instead once new serialization is implemented
        IntegerWrapper().write(io_file, 0)
        
        cur_master_loc = self.get_master_location()
        if cur_master_loc is not None:
            StringWrapper().write(io_file, str(cur_master_loc))
        else:
            StringWrapper().write(io_file, str("x"))        

        if hasattr(self, "__getstate__"):
            # The object has a user-defined serialization method.
            # Use that
            dco_extradata = self.__dclay_instance_extradata
            last_loaded_flag = dco_extradata.loaded_flag
            last_persistent_flag = dco_extradata.persistent_flag
            dco_extradata.loaded_flag = True
            dco_extradata.persistent_flag = False

            # Use pickle to the result of the serialization
            if six.PY2:
                import cPickle as pickle
            elif six.PY3:
                import _pickle as pickle
            
            state = pickle.dumps(self.__getstate__(), protocol=-1)

            # Leave the previous value, probably False & True`
            dco_extradata.loaded_flag = last_loaded_flag
            dco_extradata.persistent_flag = last_persistent_flag

            StringWrapper(mode="binary").write(io_file, state)

        else:
            # Regular dataClay provided serialization
            # Get the list of properties, making sure it is sorted
            properties = sorted(
                self.get_class_extradata().properties.values(),
                key=attrgetter('position'))

            logger.verbose("Serializing list of properties: %s", properties)

            for p in properties:

                try:
                    value = object.__getattribute__(self, "%s%s" % (DCLAY_PROPERTY_PREFIX, p.name))
                except AttributeError:
                    value = None
                
                logger.verbose("Serializing property %s with value %s ", p.name, value)

                if value is None:
                    BooleanWrapper().write(io_file, False)
                else:
                    if isinstance(p.type, UserType):
                        if not ignore_user_types:
                            BooleanWrapper().write(io_file, True)
                            SerializationLibUtilsSingleton.serialize_association(io_file, value, cur_serialized_objs, pending_objs, reference_counting)
                        else: 
                            BooleanWrapper().write(io_file, False)
                    else:
                        BooleanWrapper().write(io_file, True)
                        pck = Pickler(io_file, protocol=-1)
                        pck.persistent_id = PersistentIdPicklerHelper(cur_serialized_objs, pending_objs, reference_counting)
                        pck.dump(value)

        # Reference counting
        # TODO: this should be removed in new serialization
        # TODO: (by using paddings to directly access reference counters inside metadata)
        
        cur_stream_pos = io_file.tell()
        io_file.seek(0)
        IntegerWrapper().write(io_file, cur_stream_pos)
        io_file.seek(cur_stream_pos)        
        reference_counting.serialize_reference_counting(self.get_object_id(), io_file)

    def deserialize(self, io_file, iface_bitmaps,
                    metadata,
                    cur_deserialized_python_objs):
        """Reciprocal to serialize."""
        logger.verbose("Deserializing object %s", str(self.get_object_id()))

        # Put slow debugging info inside here:
        #
        # NOTE: new implementation of ExecutionGateway assert is not needed and wrong
        # if logger.isEnabledFor(DEBUG):
        #     klass = self.__class__
        #     logger.debug("Deserializing instance %r from class %s",
        #                  self, klass.__name__)
        #     logger.debug("The previous class is from module %s, in file %s",
        #                  klass.__module__, inspect.getfile(klass))
        #     logger.debug("The class extradata is:\n%s", klass._dclay_class_extradata)
        #     assert klass._dclay_class_extradata == self._dclay_class_extradata
        #
        # LOADED FLAG = TRUE only once deserialization is finished to avoid concurrent problems!
        # # This may be due to race conditions. It may need to do some extra locking
        # if self.__dclay_instance_extradata.loaded_flag:
        #     logger.debug("Loaded Flag is True")
        # else:
        #     self.__dclay_instance_extradata.loaded_flag = True

        """ reference counting """
        """ discard padding """ 
        IntegerWrapper().read(io_file)

        """ deserialize master_location """
        des_master_loc_str = StringWrapper().read(io_file)
        if des_master_loc_str == "x":
            self.__dclay_instance_extradata.master_location = None
        else:
            self.__dclay_instance_extradata.master_location = UUID(des_master_loc_str)

        if hasattr(self, "__setstate__"):
            # The object has a user-defined deserialization method.

            # Use pickle, and use that method instead            
            if six.PY2:
                import cPickle as pickle
            elif six.PY3:
                import _pickle as pickle
                
            state = pickle.loads(StringWrapper(mode="binary").read(io_file))
            self.__setstate__(state)

        else:
            # Regular dataClay provided deserialization

            # Start by getting the properties
            properties = sorted(
                self.get_class_extradata().properties.values(),
                key=attrgetter('position'))
            
            logger.trace("Tell io_file before loop: %s", io_file.tell())
            logger.verbose("Deserializing list of properties: %s", properties)

            for p in properties:
                
                logger.trace("Tell io_file in loop: %s", io_file.tell())
                not_null = BooleanWrapper().read(io_file)
                value = None
                if not_null:
                    logger.debug("Not null property %s", p.name)
                    if isinstance(p.type, UserType):
                        try:
                            logger.debug("Property %s is an association", p.name)
                            value = DeserializationLibUtilsSingleton.deserialize_association(io_file, iface_bitmaps, metadata, cur_deserialized_python_objs, getRuntime())
                        except KeyError as e:
                            logger.error('Failed to deserialize association', exc_info=True)
                    else:
                        try:
                            upck = Unpickler(io_file)
                            upck.persistent_load = PersistentLoadPicklerHelper(
                                metadata, cur_deserialized_python_objs, getRuntime())
                            value = upck.load()
                        except:
                            traceback.print_exc()

                logger.debug("Setting value %s for property %s", value, p.name)

                object.__setattr__(self, "%s%s" % (DCLAY_PROPERTY_PREFIX, p.name), value)
                
        """ reference counting bytes here """
        """ TODO: discard bytes? """ 

    def __reduce__(self):
        """Support for pickle protocol.

        Take into account that internal Pickle usage should be used with help
        of PersistentIdPicklerHelper and PersistentLoadPicklerHelper --for
        further information on the inner working look at the modules
        [Des|S]erializationLibUtils and both the serialize and deserialize
        methods of this class.

        This method is left here as a courtesy to end users that may need or
        want to Pickle DataClayObjects manually or through other extensions.
        """
        logger.verbose("Proceeding to `__reduce__` (Pickle-related) on a DataClayObject")
        dco_extradata = self.__dclay_instance_extradata

        if not dco_extradata.persistent_flag:
            logger.verbose("Pickling of object %r is causing a make_persistent", self)
            self.make_persistent()

        return _get_object_by_id_helper, (self.get_object_id(),
                                          self.get_class_extradata().class_id,
                                          self.get_hint())

    def __repr__(self):
        dco_extradata = self.__dclay_instance_extradata
        dcc_extradata = self.get_class_extradata()

        if dco_extradata.persistent_flag:
            return "<%s (ClassID=%s) instance with ObjectID=%s>" % \
                   (dcc_extradata.classname, dcc_extradata.class_id, dco_extradata.object_id)
        else:
            return "<%s (ClassID=%s) volatile instance with ObjectID=%s>" % \
                   (dcc_extradata.classname, dcc_extradata.class_id, dco_extradata.object_id)

    def __eq__(self, other):
        if not isinstance(other, DataClayObject):
            return False

        self_extradata = self.__dclay_instance_extradata
        other_extradata = other.__dclay_instance_extradata

        if not self_extradata.persistent_flag or not other_extradata.persistent_flag:
            return False

        return self_extradata.object_id and other_extradata.object_id \
            and self_extradata.object_id == other_extradata.object_id

    # FIXME: Think another solution, the user may want to override the method
    def __hash__(self):
        self_extradata = self.__dclay_instance_extradata
        return hash(self_extradata.object_id)

    @dclayMethod(obj='anything', property_name='str', value='anything', beforeUpdate='str', afterUpdate='str')
    def __setUpdate__(self, obj, property_name, value, beforeUpdate, afterUpdate):
        if beforeUpdate is not None:
            getattr(self, beforeUpdate)(property_name, value)
        object.__setattr__(obj, "%s%s" % ("_dataclay_property_", property_name), value)
        if afterUpdate is not None:
            getattr(self, afterUpdate)(property_name, value)




""" Class description goes here. """

'''
Created on 1 feb. 2018

@author: dgasull
'''
from dataclay.loader.DataClayObjectLoader import DataClayObjectLoader
import importlib
import traceback
import logging
import time
from dataclay.commonruntime.ExecutionGateway import ExecutionGateway
from dataclay.serialization.lib.DeserializationLibUtils import DeserializationLibUtilsSingleton
from dataclay.communication.grpc.Utils import get_metadata
from dataclay.util.classloaders.ClassLoader import load_metaclass_info
from dataclay.commonruntime.Runtime import getRuntime
from dataclay.util import Configuration


class ExecutionObjectLoader(DataClayObjectLoader):
    """
    @summary: This class is responsible to create DataClayObjects and load them with data coming from different resources. All possible
    constructions of DataClayObject should be included here. All possible "filling instance" use-cases should be managed here.
    Most lockers should be located here.
    """
    
    def __init__(self, theruntime):
        """
        @postcondition: Constructor of the object 
        @param theruntime: Runtime being managed 
        """ 
        DataClayObjectLoader.__init__(self, theruntime)
    
    def new_instance(self, class_id, object_id):
        
        self.logger.debug("Creating an instance from the class: {%s}", class_id)
    
        # Obtain the class name from the MetaClassInfo
        full_class_name, namespace = load_metaclass_info(class_id)
        self.logger.debug("MetaClassID {%s}: full class name `%s` | namespace `%s`",
                     class_id, full_class_name, namespace)
    
        class_name_parts = full_class_name.rsplit(".", 1)
    
        if len(class_name_parts) == 2:
            package_name, class_name = class_name_parts
            module_name = "%s.%s" % (namespace, package_name)
        else:
            class_name = class_name_parts[0]
            module_name = "%s" % namespace
    
        try:
            import sys
            m = importlib.import_module(module_name)
        except ImportError:
            self.logger.error("new_instance failed due to ImportError")
            self.logger.error("load_metaclass_info returned: full_class_name=%s, namespace=%s",
                         full_class_name, namespace)
            self.logger.error("Trying to import: %s", module_name)
    
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.error("DEBUG Stacktrace", exc_info=True)
    
            # # Very ugly, but required for some deep debugging strange behaviour
            # import sys
            # logger.error("The import path is: %s", sys.path)
            #
            # import subprocess
            # logger.error("`ls -laR %s` yields the following:\n%s",
            #              settings.deploy_path_source,
            #              subprocess.check_output("ls -laR %s" % settings.deploy_path_source,
            #                                      shell=True)
            #              )
    
            # Let the exception raise again, untouched
            raise
    
        klass = getattr(m, class_name)
        return klass.new_dataclay_instance(deserializing=True, object_id=object_id)
    
    def get_or_new_persistent_instance(self, class_id, object_id, hint):
        return super(ExecutionObjectLoader, self).get_or_new_persistent_instance(class_id, object_id, hint)
    
    def _get_from_db_and_fill(self, object_to_fill):
        """
        @postcondition: Get from DB and deserialize into instance
        @param object_to_fill: Instance to fill
        """
        object_id = object_to_fill.get_object_id()
        self.logger.debug("Object %s being loaded from DB", object_id)
        obj_bytes = self.runtime.get_from_sl(object_id)
        DeserializationLibUtilsSingleton.deserialize_object_from_db(object_to_fill, obj_bytes, self.runtime)
        object_to_fill.set_hint(self.runtime.get_hint())
        self.logger.debug("Object %s loaded from DB", object_id)
    
    def get_or_new_instance_from_db(self, object_id, retry):
        """
        @postcondition: Get object from memory or database and WAIT in case we are still waiting for it to be persisted.
        @param object_id: ID of the object to get
        @param retry: indicates if we should retry and wait 
        @param class_id: Can be none. Class ID of the object to get. In order to avoid looking for metadata.
        @return: the object 
        """
        
        """
        Retry while object is not 'registered' (not talking about 'stored'!)
        IMPORTANT: This is different than waiting for an object to be stored
        If the object is not registered we still do not know the class id of the instance
        in which to load the bytes.
        Due to concurrency we should read bytes and deserialize and unlock.
        Therefore there is Two waiting loops. (can we do it better?, more locking?)
        """ 
        self.logger.verbose("Get or create new instance from SL with object id %s in Heap ", str(object_id))
        obtained = False 
        wait_time = 0
        sleep_time = Configuration.SLEEP_WAIT_REGISTERED / 1000
        instance = None
        while not obtained:
            self.runtime.lock(object_id)
            try:
                instance = self.runtime.get_from_heap(object_id)
                if instance is None: 
                    obj_bytes = self.runtime.get_from_sl(object_id)
                    msg = DeserializationLibUtilsSingleton.deserialize_grpc_message_from_db(obj_bytes)
                    metadata = get_metadata(msg.metadata)
                    instance_class_id = metadata[1][0]
                    instance = self.new_instance(instance_class_id, object_id)
                    instance.initialize_object_as_persistent()
                    DeserializationLibUtilsSingleton.deserialize_object_from_db_bytes_aux(instance, metadata, msg.data, self.runtime)
                    instance.set_hint(self.runtime.get_hint())
                    self.logger.debug("Object %s deserialized", object_id)

                if not instance.is_loaded():
                    self._get_from_db_and_fill(instance)

                obtained = True
            except:
                self.logger.debug("Received error while retrieving object %s", object_id, exc_info=True)
                if not retry or wait_time > Configuration.TIMEOUT_WAIT_REGISTERED:
                    raise
                
                wait_time = wait_time + sleep_time
                self.logger.debug("Object %s not found in DB. Waiting and retry...", object_id)
                time.sleep(sleep_time)
            finally:
                self.runtime.unlock(object_id)
        
        return instance
                
    def load_object_from_db(self, instance, retry):
        """
        @postcondition: Load DataClayObject from Database
        @param instance: DataClayObject instance to fill
        @param retry: Indicates retry loading in case it is not in db.
        """
        
        object_id = instance.get_object_id()
        loaded = False 
        wait_time = 0
        sleep_time = Configuration.SLEEP_WAIT_REGISTERED / 1000
        while not loaded and wait_time < Configuration.TIMEOUT_WAIT_REGISTERED:
            self.runtime.lock(object_id)
            try:
                """ double check for race-conditions"""
                if not instance.is_loaded():
                    self._get_from_db_and_fill(instance)
                loaded = True
            except Exception as ex: 
                traceback.print_exc()
                if not retry or wait_time > Configuration.TIMEOUT_WAIT_REGISTERED:
                    raise ex
                self.logger.debug("Object %s not found in DB. Waiting and retry...", object_id)
                wait_time = wait_time + sleep_time
                time.sleep(sleep_time)
            finally:
                self.runtime.unlock(object_id)
                
    def get_or_new_volatile_instance_and_load(self, class_id, object_id, hint, obj_with_data, ifacebitmaps):
        """
        @postcondition: Get from Heap or create a new volatile in EE and load data on it.
        @param class_id: id of the class of the object
        @param object_id: id of the object
        @param hint: hint of the object 
        @param obj_with_data: data of the volatile 
        @param ifacebitmaps: interface bitmaps 
        """
        """
        RACE CONDITION DESIGN
        There are two objects A and B, A -> B, A is persistent and B is volatile.
        There are two threads T1 and T2, T1 is executing a method on A that uses B, when deserializing A, B is loaded into
        heap as a persistent object (all associations are persistent). However, it is actually a volatile send by T2.
        When a volatile server is received and a persistent instance is found, this persistent instance should be "replaced"
        by the new volatile server.
        """
        self.runtime.lock(object_id)
        try:
            """ Double check for race conditions """
            self.logger.verbose("Get or create new volatile instance in EE with object id %s in Heap ", str(object_id))
            volatile_obj = self.runtime.get_from_heap(object_id)
            if volatile_obj is None: 
                volatile_obj = self.new_instance_internal(class_id, object_id, hint)
                    
            """ Deserialize volatile """ 
            DeserializationLibUtilsSingleton.deserialize_object_with_data(obj_with_data, volatile_obj, ifacebitmaps, self.runtime,
                                             self.runtime.get_session_id(),
                                             True)
            
            self.logger.debug("Object %s deserialized", object_id)

            # WARNING: RACE CONDITION at EE - during deserialization of volatiles the
            # object may be created and
            # loaded in Heap but not "fully deserialized" yet so even if any execution find
            # it in the
            # heap, object might
            # be not ready (null fields, and no, so is loaded cannot
            # be true till object was fully deserialized)
            volatile_obj.initialize_object_as_volatile()
        finally:
            self.runtime.unlock(object_id)
        
        return volatile_obj
        




""" Class description goes here. """

"""Common Python package.

Note that the client "entrypoint" is the package dataclay.api. Importing it
ensures proper initialization of client-side stuff.

The Execution Environment should take care of the RuntimeType and then call the
dataclay.core.initialize function.
"""
import threading
import logging

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)

__all__ = ['DataClayObject', 'StorageObject', 'dclayMethod']

""" Local variables in thread """
threadLocal = threading.local() 

""" Cache of runtimes per thread """
runtimes_per_thread = dict()  

""" Runtime for all threads in client side. It's a singleton, created once. """
client_static_runtime = None


def clean_runtime():
    global threadLocal
    global runtimes_per_thread
    global client_static_runtime
    threadLocal = threading.local()
    runtimes_per_thread.clear()
    client_static_runtime = None


def setRuntime(runtime):
    """
    @summary: set runtime in current thread. This function is used in Execution Environment 
    @param runtime: runtime to set 
    """
    thread_id = threading.current_thread().ident
    runtimes_per_thread[thread_id] = runtime


def getRuntime():
    """
    @summary: get runtime associated to current thread. If there is no runtime, return client runtime. 
    In EE if current thread has no runtime, it means thread was not prepared properly and it is wrong!
    @return runtime of the current thread
    """
    
    global client_static_runtime
    thread_id = threading.current_thread().ident
    if thread_id in runtimes_per_thread:
        # logger.info("==== Welcome back thread %s", threading.current_thread().name)
        return runtimes_per_thread.get(thread_id)
    else:
        if client_static_runtime is None:
            from dataclay.commonruntime.ClientRuntime import ClientRuntime
            client_static_runtime = ClientRuntime()
        return client_static_runtime


# Those need the commonruntime
from dataclay.DataClayObject import DataClayObject  # Import after runtime to avoid circular-imports (runtime is already defined here)
StorageObject = DataClayObject  # Redundant alias





""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class VersionInfo(ManagementObject):
    _fields = ["versionOID",
               "versionsMapping",
               "locID",
               "originalMD",
               ]




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class Account(ManagementObject):
    _fields = ["username",
               "credential",
               "role",
               ]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dataclay/communication/grpc/generated/dataservice/dataservice.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from dataclay.communication.grpc.messages.dataservice import dataservice_messages_pb2 as dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2
from dataclay.communication.grpc.messages.common import common_messages_pb2 as dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='dataclay/communication/grpc/generated/dataservice/dataservice.proto',
  package='dataclay.communication.grpc.dataservice',
  syntax='proto3',
  serialized_options=_b('\n1dataclay.communication.grpc.generated.dataserviceB\026DataServiceGrpcService'),
  serialized_pb=_b('\nCdataclay/communication/grpc/generated/dataservice/dataservice.proto\x12\'dataclay.communication.grpc.dataservice\x1aKdataclay/communication/grpc/messages/dataservice/dataservice_messages.proto\x1a\x41\x64\x61taclay/communication/grpc/messages/common/common_messages.proto2\xf2/\n\x0b\x44\x61taService\x12\x83\x01\n\rinitBackendID\x12=.dataclay.communication.grpc.dataservice.InitBackendIDRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\xa3\x01\n\x1d\x61ssociateExecutionEnvironment\x12M.dataclay.communication.grpc.dataservice.AssociateExecutionEnvironmentRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x8b\x01\n\x11\x64\x65ployMetaClasses\x12\x41.dataclay.communication.grpc.dataservice.DeployMetaClassesRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x83\x01\n\rdeployClasses\x12=.dataclay.communication.grpc.dataservice.DeployClassesRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x7f\n\x0b\x65nrichClass\x12;.dataclay.communication.grpc.dataservice.EnrichClassRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\xa8\x01\n\x15newPersistentInstance\x12\x45.dataclay.communication.grpc.dataservice.NewPersistentInstanceRequest\x1a\x46.dataclay.communication.grpc.dataservice.NewPersistentInstanceResponse\"\x00\x12\x81\x01\n\x0cstoreObjects\x12<.dataclay.communication.grpc.dataservice.StoreObjectsRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x96\x01\n\x0fgetCopyOfObject\x12?.dataclay.communication.grpc.dataservice.GetCopyOfObjectRequest\x1a@.dataclay.communication.grpc.dataservice.GetCopyOfObjectResponse\"\x00\x12\x81\x01\n\x0cupdateObject\x12<.dataclay.communication.grpc.dataservice.UpdateObjectRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x87\x01\n\ngetObjects\x12:.dataclay.communication.grpc.dataservice.GetObjectsRequest\x1a;.dataclay.communication.grpc.dataservice.GetObjectsResponse\"\x00\x12\x7f\n\x0bnewMetaData\x12;.dataclay.communication.grpc.dataservice.NewMetaDataRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x87\x01\n\nnewVersion\x12:.dataclay.communication.grpc.dataservice.NewVersionRequest\x1a;.dataclay.communication.grpc.dataservice.NewVersionResponse\"\x00\x12\x8d\x01\n\x12\x63onsolidateVersion\x12\x42.dataclay.communication.grpc.dataservice.ConsolidateVersionRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x83\x01\n\rupsertObjects\x12=.dataclay.communication.grpc.dataservice.UpsertObjectsRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x87\x01\n\nnewReplica\x12:.dataclay.communication.grpc.dataservice.NewReplicaRequest\x1a;.dataclay.communication.grpc.dataservice.NewReplicaResponse\"\x00\x12\x8a\x01\n\x0bmoveObjects\x12;.dataclay.communication.grpc.dataservice.MoveObjectsRequest\x1a<.dataclay.communication.grpc.dataservice.MoveObjectsResponse\"\x00\x12\x90\x01\n\rremoveObjects\x12=.dataclay.communication.grpc.dataservice.RemoveObjectsRequest\x1a>.dataclay.communication.grpc.dataservice.RemoveObjectsResponse\"\x00\x12\x9d\x01\n\x18migrateObjectsToBackends\x12>.dataclay.communication.grpc.dataservice.MigrateObjectsRequest\x1a?.dataclay.communication.grpc.dataservice.MigrateObjectsResponse\"\x00\x12\xbd\x01\n\x1cgetClassIDFromObjectInMemory\x12L.dataclay.communication.grpc.dataservice.GetClassIDFromObjectInMemoryRequest\x1aM.dataclay.communication.grpc.dataservice.GetClassIDFromObjectInMemoryResponse\"\x00\x12\xa8\x01\n\x15\x65xecuteImplementation\x12\x45.dataclay.communication.grpc.dataservice.ExecuteImplementationRequest\x1a\x46.dataclay.communication.grpc.dataservice.ExecuteImplementationResponse\"\x00\x12\x85\x01\n\x0emakePersistent\x12>.dataclay.communication.grpc.dataservice.MakePersistentRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12y\n\x08\x66\x65\x64\x65rate\x12\x38.dataclay.communication.grpc.dataservice.FederateRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12}\n\nunfederate\x12:.dataclay.communication.grpc.dataservice.UnfederateRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12{\n\x06\x65xists\x12\x36.dataclay.communication.grpc.dataservice.ExistsRequest\x1a\x37.dataclay.communication.grpc.dataservice.ExistsResponse\"\x00\x12\xa2\x01\n\x13getFederatedObjects\x12\x43.dataclay.communication.grpc.dataservice.GetFederatedObjectsRequest\x1a\x44.dataclay.communication.grpc.dataservice.GetFederatedObjectsResponse\"\x00\x12\xac\x01\n\x17getReferencedObjectsIDs\x12\x46.dataclay.communication.grpc.dataservice.GetReferencedObjectIDsRequest\x1aG.dataclay.communication.grpc.dataservice.GetReferencedObjectIDsResponse\"\x00\x12\x8d\x01\n\x0c\x66ilterObject\x12<.dataclay.communication.grpc.dataservice.FilterObjectRequest\x1a=.dataclay.communication.grpc.dataservice.FilterObjectResponse\"\x00\x12{\n\tstoreToDB\x12\x39.dataclay.communication.grpc.dataservice.StoreToDBRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x84\x01\n\tgetFromDB\x12\x39.dataclay.communication.grpc.dataservice.GetFromDBRequest\x1a:.dataclay.communication.grpc.dataservice.GetFromDBResponse\"\x00\x12}\n\nupdateToDB\x12:.dataclay.communication.grpc.dataservice.UpdateToDBRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12}\n\ndeleteToDB\x12:.dataclay.communication.grpc.dataservice.DeleteToDBRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x87\x01\n\nexistsInDB\x12:.dataclay.communication.grpc.dataservice.ExistsInDBRequest\x1a;.dataclay.communication.grpc.dataservice.ExistsInDBResponse\"\x00\x12\x85\x01\n\x1c\x63leanExecutionClassDirectory\x12\x30.dataclay.communication.grpc.common.EmptyMessage\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12w\n\x0e\x63loseDbHandler\x12\x30.dataclay.communication.grpc.common.EmptyMessage\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12q\n\x08shutDown\x12\x30.dataclay.communication.grpc.common.EmptyMessage\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12}\n\x14\x64isconnectFromOthers\x12\x30.dataclay.communication.grpc.common.EmptyMessage\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x7f\n\x16registerPendingObjects\x12\x30.dataclay.communication.grpc.common.EmptyMessage\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12t\n\x0b\x63leanCaches\x12\x30.dataclay.communication.grpc.common.EmptyMessage\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x87\x01\n\x0f\x61\x63tivateTracing\x12?.dataclay.communication.grpc.dataservice.ActivateTracingRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12z\n\x11\x64\x65\x61\x63tivateTracing\x12\x30.dataclay.communication.grpc.common.EmptyMessage\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12v\n\tgetTraces\x12\x30.dataclay.communication.grpc.common.EmptyMessage\x1a\x35.dataclay.communication.grpc.common.GetTracesResponse\"\x00\x12\x89\x01\n\x10\x63loseSessionInDS\x12@.dataclay.communication.grpc.dataservice.CloseSessionInDSRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12}\n\nupdateRefs\x12:.dataclay.communication.grpc.dataservice.UpdateRefsRequest\x1a\x31.dataclay.communication.grpc.common.ExceptionInfo\"\x00\x12\x93\x01\n\x15getRetainedReferences\x12\x30.dataclay.communication.grpc.common.EmptyMessage\x1a\x46.dataclay.communication.grpc.dataservice.GetRetainedReferencesResponse\"\x00\x42K\n1dataclay.communication.grpc.generated.dataserviceB\x16\x44\x61taServiceGrpcServiceb\x06proto3')
  ,
  dependencies=[dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2.DESCRIPTOR,dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_DATASERVICE = _descriptor.ServiceDescriptor(
  name='DataService',
  full_name='dataclay.communication.grpc.dataservice.DataService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=257,
  serialized_end=6387,
  methods=[
  _descriptor.MethodDescriptor(
    name='initBackendID',
    full_name='dataclay.communication.grpc.dataservice.DataService.initBackendID',
    index=0,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._INITBACKENDIDREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='associateExecutionEnvironment',
    full_name='dataclay.communication.grpc.dataservice.DataService.associateExecutionEnvironment',
    index=1,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._ASSOCIATEEXECUTIONENVIRONMENTREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='deployMetaClasses',
    full_name='dataclay.communication.grpc.dataservice.DataService.deployMetaClasses',
    index=2,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._DEPLOYMETACLASSESREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='deployClasses',
    full_name='dataclay.communication.grpc.dataservice.DataService.deployClasses',
    index=3,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._DEPLOYCLASSESREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='enrichClass',
    full_name='dataclay.communication.grpc.dataservice.DataService.enrichClass',
    index=4,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._ENRICHCLASSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='newPersistentInstance',
    full_name='dataclay.communication.grpc.dataservice.DataService.newPersistentInstance',
    index=5,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._NEWPERSISTENTINSTANCEREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._NEWPERSISTENTINSTANCERESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='storeObjects',
    full_name='dataclay.communication.grpc.dataservice.DataService.storeObjects',
    index=6,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._STOREOBJECTSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='getCopyOfObject',
    full_name='dataclay.communication.grpc.dataservice.DataService.getCopyOfObject',
    index=7,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETCOPYOFOBJECTREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETCOPYOFOBJECTRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='updateObject',
    full_name='dataclay.communication.grpc.dataservice.DataService.updateObject',
    index=8,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._UPDATEOBJECTREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='getObjects',
    full_name='dataclay.communication.grpc.dataservice.DataService.getObjects',
    index=9,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETOBJECTSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETOBJECTSRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='newMetaData',
    full_name='dataclay.communication.grpc.dataservice.DataService.newMetaData',
    index=10,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._NEWMETADATAREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='newVersion',
    full_name='dataclay.communication.grpc.dataservice.DataService.newVersion',
    index=11,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._NEWVERSIONREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._NEWVERSIONRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='consolidateVersion',
    full_name='dataclay.communication.grpc.dataservice.DataService.consolidateVersion',
    index=12,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._CONSOLIDATEVERSIONREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='upsertObjects',
    full_name='dataclay.communication.grpc.dataservice.DataService.upsertObjects',
    index=13,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._UPSERTOBJECTSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='newReplica',
    full_name='dataclay.communication.grpc.dataservice.DataService.newReplica',
    index=14,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._NEWREPLICAREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._NEWREPLICARESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='moveObjects',
    full_name='dataclay.communication.grpc.dataservice.DataService.moveObjects',
    index=15,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._MOVEOBJECTSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._MOVEOBJECTSRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='removeObjects',
    full_name='dataclay.communication.grpc.dataservice.DataService.removeObjects',
    index=16,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._REMOVEOBJECTSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._REMOVEOBJECTSRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='migrateObjectsToBackends',
    full_name='dataclay.communication.grpc.dataservice.DataService.migrateObjectsToBackends',
    index=17,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._MIGRATEOBJECTSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._MIGRATEOBJECTSRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='getClassIDFromObjectInMemory',
    full_name='dataclay.communication.grpc.dataservice.DataService.getClassIDFromObjectInMemory',
    index=18,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETCLASSIDFROMOBJECTINMEMORYREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETCLASSIDFROMOBJECTINMEMORYRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='executeImplementation',
    full_name='dataclay.communication.grpc.dataservice.DataService.executeImplementation',
    index=19,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._EXECUTEIMPLEMENTATIONREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._EXECUTEIMPLEMENTATIONRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='makePersistent',
    full_name='dataclay.communication.grpc.dataservice.DataService.makePersistent',
    index=20,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._MAKEPERSISTENTREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='federate',
    full_name='dataclay.communication.grpc.dataservice.DataService.federate',
    index=21,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._FEDERATEREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='unfederate',
    full_name='dataclay.communication.grpc.dataservice.DataService.unfederate',
    index=22,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._UNFEDERATEREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='exists',
    full_name='dataclay.communication.grpc.dataservice.DataService.exists',
    index=23,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._EXISTSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._EXISTSRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='getFederatedObjects',
    full_name='dataclay.communication.grpc.dataservice.DataService.getFederatedObjects',
    index=24,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETFEDERATEDOBJECTSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETFEDERATEDOBJECTSRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='getReferencedObjectsIDs',
    full_name='dataclay.communication.grpc.dataservice.DataService.getReferencedObjectsIDs',
    index=25,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETREFERENCEDOBJECTIDSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETREFERENCEDOBJECTIDSRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='filterObject',
    full_name='dataclay.communication.grpc.dataservice.DataService.filterObject',
    index=26,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._FILTEROBJECTREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._FILTEROBJECTRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='storeToDB',
    full_name='dataclay.communication.grpc.dataservice.DataService.storeToDB',
    index=27,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._STORETODBREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='getFromDB',
    full_name='dataclay.communication.grpc.dataservice.DataService.getFromDB',
    index=28,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETFROMDBREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETFROMDBRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='updateToDB',
    full_name='dataclay.communication.grpc.dataservice.DataService.updateToDB',
    index=29,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._UPDATETODBREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='deleteToDB',
    full_name='dataclay.communication.grpc.dataservice.DataService.deleteToDB',
    index=30,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._DELETETODBREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='existsInDB',
    full_name='dataclay.communication.grpc.dataservice.DataService.existsInDB',
    index=31,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._EXISTSINDBREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._EXISTSINDBRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='cleanExecutionClassDirectory',
    full_name='dataclay.communication.grpc.dataservice.DataService.cleanExecutionClassDirectory',
    index=32,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EMPTYMESSAGE,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='closeDbHandler',
    full_name='dataclay.communication.grpc.dataservice.DataService.closeDbHandler',
    index=33,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EMPTYMESSAGE,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='shutDown',
    full_name='dataclay.communication.grpc.dataservice.DataService.shutDown',
    index=34,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EMPTYMESSAGE,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='disconnectFromOthers',
    full_name='dataclay.communication.grpc.dataservice.DataService.disconnectFromOthers',
    index=35,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EMPTYMESSAGE,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='registerPendingObjects',
    full_name='dataclay.communication.grpc.dataservice.DataService.registerPendingObjects',
    index=36,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EMPTYMESSAGE,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='cleanCaches',
    full_name='dataclay.communication.grpc.dataservice.DataService.cleanCaches',
    index=37,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EMPTYMESSAGE,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='activateTracing',
    full_name='dataclay.communication.grpc.dataservice.DataService.activateTracing',
    index=38,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._ACTIVATETRACINGREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='deactivateTracing',
    full_name='dataclay.communication.grpc.dataservice.DataService.deactivateTracing',
    index=39,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EMPTYMESSAGE,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='getTraces',
    full_name='dataclay.communication.grpc.dataservice.DataService.getTraces',
    index=40,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EMPTYMESSAGE,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._GETTRACESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='closeSessionInDS',
    full_name='dataclay.communication.grpc.dataservice.DataService.closeSessionInDS',
    index=41,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._CLOSESESSIONINDSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='updateRefs',
    full_name='dataclay.communication.grpc.dataservice.DataService.updateRefs',
    index=42,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._UPDATEREFSREQUEST,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EXCEPTIONINFO,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='getRetainedReferences',
    full_name='dataclay.communication.grpc.dataservice.DataService.getRetainedReferences',
    index=43,
    containing_service=None,
    input_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_common_dot_common__messages__pb2._EMPTYMESSAGE,
    output_type=dataclay_dot_communication_dot_grpc_dot_messages_dot_dataservice_dot_dataservice__messages__pb2._GETRETAINEDREFERENCESRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_DATASERVICE)

DESCRIPTOR.services_by_name['DataService'] = _DATASERVICE

# @@protoc_insertion_point(module_scope)




""" Class description goes here. """

"""Deserialization code related to DataClay objects

The code implemented in this module is (at "this moment") identical to the ones
implemented in the Java package client.CommonLib. Specifically, the deserialize*
functions are more or less adapted here.
"""

from io import BytesIO
import logging

from dataclay.exceptions.exceptions import InvalidPythonSignature
from dataclay.serialization.python.lang.IntegerWrapper import IntegerWrapper
from dataclay.serialization.python.lang.VLQIntegerWrapper import VLQIntegerWrapper
from dataclay.serialization.python.util.PyTypeWildcardWrapper import PyTypeWildcardWrapper
import dataclay.communication.grpc.messages.common.common_messages_pb2 as common_messages
from dataclay.communication.grpc.Utils import get_metadata

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)

class DeserializationLibUtils(object):
    def deserialize_params(self, serialized_params_or_return, iface_bitmaps,
            param_specs, params_order, runtime):
        return self.deserialize_params_or_return(serialized_params_or_return,
                            iface_bitmaps,
                            param_specs,
                            params_order, runtime)


    def _create_buffer_and_deserialize(self, obj_bytes, instance, ifacebitmaps, metadata, cur_deser_python_objs):
        """
        @postcondition: Create buffer and deserialize
        @param obj_bytes: bytes to deserialize
        @param instance: Instance to deserialize
        @param ifacebitmaps: Map of bitmaps representing the interfaces to use
        @param metadata: object metadata
        @param cur_deser_python_objs: Currently deserialized Java objects
        """

        # TODO: obj_bytes is an io_file, use a buffer instead
        instance.deserialize(obj_bytes, ifacebitmaps, metadata, cur_deser_python_objs)


    def deserialize_grpc_message_from_db(self, obj_bytes):
        """
        @postcondition: Deserialize grpc message from DB
        @param obj_bytes: object bytes
        @return grpc msg stored in db
        """
        msg = common_messages.PersistentObjectInDB()
        msg.ParseFromString(obj_bytes)
        return msg


    def deserialize_object_from_db(self, object_to_fill, object_bytes, runtime):
        """
        @postcondition: Deserialize object from bytes
        @param object_to_fill: object to fill
        @param object_bytes: Object bytes
        @param runtime: runtime
        """

        msg = self.deserialize_grpc_message_from_db(object_bytes)
        metadata = get_metadata(msg.metadata)
        self.deserialize_object_from_db_bytes_aux(object_to_fill, metadata, msg.data, runtime)


    def deserialize_object_from_db_bytes_aux(self, object_to_fill, metadata, data, runtime):
        """
        @postcondition: Deserialize object from bytes into instance provided
        @param metadata: object metadata
        @param data: Object data
        @param runtime: runtime
        """

        cur_deser_python_objs = dict()
        obj_bytes = BytesIO(data)
        object_to_fill.set_loaded(True)
        object_to_fill.set_persistent(True)
        self._create_buffer_and_deserialize(obj_bytes, object_to_fill, None, metadata, cur_deser_python_objs)

        """
        GARBAGE COLLECTOR RACE CONDITION
        ================================
        If some object was cleaned and removed from GC retained refs, it does NOT mean it was removed
        from Weak references map (Heap) because we will ONLY remove an entry in Heap if it is
        pointing to NULL, i.e. the Java GC removed it.
        So, if some execution is requested after we remove an entry from retained refs (we cleaned and send
        the object to disk), we check if the
        object is in Heap (see executeImplementation in DS) and therefore, we created a new reference
        making impossible for Java GC to clean the reference. We will add the object to retained references
        again once it is deserialized from DB.
        """
        runtime.add_to_heap(object_to_fill)

    def deserialize_objbytes_from_db_into_obj_data(self, object_id, obj_bytes_from_db, runtime):
        """
        @postcondition: Deserialize object from bytes into ObjectData and return it.
        @param object_id: ID of object
        @param obj_bytes_from_db: Object bytes
        @param runtime: Runtime to use
        @return object data (to send) of this object
        """
        logger.debug("OBJECT_ID %s, OBJ_BYTES %s", object_id, obj_bytes_from_db)

        msg = self.deserialize_grpc_message_from_db(obj_bytes_from_db)
        metadata = get_metadata(msg.metadata)
        obj_bytes = BytesIO(msg.data)

        logger.debug("METADATA %s and METADATA CLASS_ID %s", metadata, metadata[1][0])

        return object_id, metadata[1][0], metadata, obj_bytes


    def deserialize_metadata_from_db(self, obj_bytes_from_db):
        """
        @postcondition: deserialize metadata from db (just metadata)
        @param obj_bytes_from_db: object bytes from db
        @return metadata of the object
        """
        msg = common_messages.PersistentObjectInDB()
        msg.ParseFromString(obj_bytes_from_db)
        metadata = get_metadata(msg.metadata)
        return metadata


    def deserialize_object_with_data_in_client(self, param_or_ret, instance, ifacebitmpas, runtime, owner_session_id):
        """
        @postcondition: Deserialize object into a memory instance. ALSO called from executeImplementation in case of 'executions during
        deserialization'. THIS FUNCTION SHOULD NEVER BE CALLED FROM CLIENT SIDE.
        @param param_or_ret: Param/return bytes and metadata
        @param instance: Object in which to deserialize data
        @param ifacebitmaps: Interface bitmaps
        @param runtime: the runtime
        @param owner_session_id: Can be None. ID of owner session of the object. Used for volatiles pending to register.
        """

        """ Lock object until deserialization is finished in case another instance is waiting to do the same so much """
        runtime.lock(instance.get_object_id())
        try:
            metadata = param_or_ret[2]
            obj_bytes = BytesIO(param_or_ret[3])
            cur_deser_python_objs = dict()
            self._create_buffer_and_deserialize(obj_bytes, instance, None, metadata, cur_deser_python_objs)
            instance.set_persistent(False)
            instance.set_hint(None)

        finally:
            runtime.unlock(instance.get_object_id())


    def deserialize_object_with_data(self, param_or_ret, instance, ifacebitmpas, runtime, owner_session_id, force_deserialization):
        """
        @postcondition: Deserialize object into a memory instance. ALSO called from executeImplementation in case of 'executions during
        deserialization'. THIS FUNCTION SHOULD NEVER BE CALLED FROM CLIENT SIDE.
        @param param_or_ret: Param/return bytes and metadata
        @param instance: Object in which to deserialize data
        @param ifacebitmaps: Interface bitmaps
        @param runtime: the runtime
        @param owner_session_id: Can be None. ID of owner session of the object. Used for volatiles pending to register.
        @param force_deserialization: Check if the object is loaded or not. If FALSE and loaded, then no deserialization is happening. If TRUE, then
        deserialization is forced.
        """

        """
        FORCE DESERIALIZATION NOTE.
        Volatiles in server can be created in two ways: received from client or in some execution.
        If created during execution, flag loaded = true, but if received, flag loaded should be false until
        the object is properly loaded. In order to control the first case, constructor of the object is
        setting loaded to true and here we set it to false. Therefore we need to "force" deserialization sometimes.
        Also happening in update of objects and cleaning in GC.
        """
        """ Lock object until deserialization is finished in case another instance is waiting to do the same so much """
        runtime.lock(instance.get_object_id())
        try:
            if force_deserialization or not instance.is_loaded():
                """ TODO: improve GRPC messages """
                metadata = param_or_ret[2]
                obj_bytes = BytesIO(param_or_ret[3])
                cur_deser_python_objs = dict()
                self._create_buffer_and_deserialize(obj_bytes, instance, None, metadata, cur_deser_python_objs)
                instance.set_loaded(True)
                instance.set_persistent(True)
                if owner_session_id is not None:
                    instance.set_owner_session_id(owner_session_id)

        finally:
            runtime.unlock(instance.get_object_id())


    def deserialize_return(self, serialized_params_or_return, iface_bitmaps, return_type, runtime):

        if serialized_params_or_return[0] == 0:
            logger.verbose("No return to deserialize: returning None")
            return None
        return self.deserialize_params_or_return(serialized_params_or_return,
                            iface_bitmaps,
                            {"0": return_type},
                            ["0"], runtime)[0]


    def deserialize_params_or_return(self, serialized_params_or_return,
            iface_bitmaps, param_specs, params_order, runtime):
        """
        Deserialize parameters or return of an execution
        :param serialized_params_or_return: serialized parameters or return
        :param iface_bitmaps: interface bitmaps
        :param param_specs: specifications of parameters
        :param params_order: order of parameters
        :param runtime: runtime being used
        """
        num_params = serialized_params_or_return[0]
        params = [None] * num_params

        """ VOLATILES """
        first_volatile = True
        for i, serialized_param in serialized_params_or_return[3].items():
            object_id = serialized_param[0]
            class_id = serialized_param[1]
            logger.verbose("Deserializing volatile with object ID %s" % str(object_id))

            if first_volatile:
                runtime.add_volatiles_under_deserialization(serialized_params_or_return[3])
                first_volatile = False

            deserialized_param = runtime.get_or_new_volatile_instance_and_load(object_id, class_id, runtime.get_hint(),
                                                                               serialized_param, iface_bitmaps)

            # For makePersistent or federate methods, we must know that there is a session using the object which
            # is the one that persisted/federated. Normally, we know a client is using an object if we serialize
            # it to send to him but in that case client has created/federated the object.
            runtime.add_session_reference(deserialized_param.get_object_id())

            if i < num_params:
                params[i] = deserialized_param

        """ LANGUAGE """
        for i, serialized_param in serialized_params_or_return[2].items():
            logger.verbose("Deserializing language type at index %s" % str(i))

            obj_bytes = BytesIO(serialized_param[1])
            params[i] = self.deserialize_language(obj_bytes, param_specs[params_order[i]])

        """ IMMUTABLES """
        for i, serialized_param in serialized_params_or_return[1].items():
            logger.verbose("Deserializing immutable type at index %s" % str(i))

            obj_bytes = BytesIO(serialized_param)
            params[i] = self.deserialize_immutable(obj_bytes, param_specs[params_order[i]])

        """ PERSISTENT """
        for i, serialized_param in serialized_params_or_return[4].items():

            object_id = serialized_param[0]
            hint = serialized_param[1]
            class_id = serialized_param[2]
            logger.verbose("Deserializing persistent object with object ID %s" % str(object_id))

            deserialized_param = runtime.get_or_new_persistent_instance(object_id, class_id, hint)
            deserialized_param.set_persistent(True)
            if i < num_params:
                params[i] = deserialized_param
                # Set hint in metadata (ToDo: create general getOrNewInstance)

        if not first_volatile:
            runtime.remove_volatiles_under_deserialization()

        return params


    def deserialize_association(self, io_file, iface_bitmaps,
            metadata, cur_deserialized_objs, runtime):
        """
        @postcondition: deserialize association
        @param io_file: bytes of the object containing the association
        @param iface_bitmaps: interface bitmaps
        @param metadata: metadata of the object
        @param cur_deserialized_objs: current deserialized objects
        @param runtime: the runtime
        """

        tag = VLQIntegerWrapper().read(io_file)
        logger.debug("Deserializing association for tag: %d", tag)
        logger.debug("Metadata OIDs: %s", metadata[0])
        object_id = metadata[0][tag]
        logger.debug("Metadata ClassIDs: %s", metadata[1])
        metaclass_id = metadata[1][tag]
        # TODO: NumRefs ( metadata[3]) not deserialized?? Ask about
        # How/Where use NumRefs (metadata[3])?

        hint = None
        try:
            logger.info("Metadata Hints: %s", metadata[2])
            hint = metadata[2][tag]
        except KeyError:
            pass
        logger.debug("Deserializing association to object: %s", str(object_id))
        obj = runtime.get_or_new_persistent_instance(object_id, metaclass_id, hint)
        cur_deserialized_objs[tag] = obj
        return obj


    def deserialize_immutable(self, io_file, type_):
        # Well, if it is a primitive type we already have them...
        try:
            ptw = PyTypeWildcardWrapper(type_.signature)
        except InvalidPythonSignature:
            raise NotImplementedError("Only Java primitive types are "
                                          "understood in Python.")
        return ptw.read(io_file)


    def deserialize_language(self, io_file, type_):
        # Well, if it is a primitive type we already have them...
        try:
            ptw = PyTypeWildcardWrapper(type_.signature)
        except InvalidPythonSignature:
            raise NotImplementedError("In fact, InvalidPythonSignature was "
                                          "not even implemented, seems somebody is "
                                          "raising it without implementing logic.")
        return ptw.read(io_file)


    def extract_reference_counting(self, io_file):
        io_file.seek(0)
        ref_counting_pos = IntegerWrapper().read(io_file)
        io_file.seek(ref_counting_pos)
        # read up to last byte
        return io_file.read()

DeserializationLibUtilsSingleton = DeserializationLibUtils()

class PersistentLoadPicklerHelper(object):
    """Helper to solve deserialization of associations inside Pickled structures.

    See https://docs.python.org/2.7/library/pickle.html#pickling-and-unpickling-external-objects
    for more information of what this is doing.

    The `__call__` method is being called by Pickle for persistent ids, i.e.,
    objects that have been serialized by PersistentIdPicklerHelper and thus
    are associations (DataClayObjects).

    The code is very close to the deserialize association.
    """

    def __init__(self, metadata, cur_deserialized_objs, runtime):
        self._metadata = metadata
        self._runtime = runtime
        self._cur_deserialized_objs = cur_deserialized_objs

    def __call__(self, str_tag):
        tag = int(str_tag)
        logger.verbose("Deserializing association for tag: %d", tag)
        logger.debug("Metadata OIDs: %s", self._metadata[0])
        object_id = self._metadata[0][tag]
        logger.debug("Metadata ClassIDs: %s", self._metadata[1])
        metaclass_id = self._metadata[1][tag]

        # TODO: NumRefs ( metadata[3]) not deserialized?? Ask about
        # How/Where use NumRefs (metadata[3])?

        hint = None
        try:
            logger.debug("Metadata Hints: %s", self._metadata[2])
            hint = self._metadata[2][tag]
        except KeyError:
            logger.debug("No Metadata Hints")
            pass

        logger.debug("Deserializing association to object: %s", str(object_id))
        obj = self._runtime.get_or_new_persistent_instance(object_id, metaclass_id, hint)
        self._cur_deserialized_objs[tag] = obj
        return obj




""" Class description goes here. """

from dataclay import StorageObject
from dataclay import dclayMethod
import logging

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2016 Barcelona Supercomputing Center (BSC-CNS)'

CLASSES_TO_REGISTER = (
    "GenericSplit", "WorkStealingSplit", "WorkMovingSplit",
)
logger = logging.getLogger(__name__)


def split(iterable, **split_options):
    """Perform a split on iterable.

    This method is highly inspired in the `iter` global method (in conjunction
    with its __iter__ counterpart method) for iterable classes.

    :param iterable: An iterable, which will typically be a Storage<Collection>
    :param split_options: The optional additional arguments to the split method.
    May be ignored.
    :return: A collection of Split, or something similar. If iterable is not a
    Storage<Collection>, returns a tuple with a single element, the iterable
    argument itself
    """
    try:
        # Default behaviour is to use the data model `split` method
        return iterable.split(**split_options)
    except AttributeError:
        # Otherwise, simply return a iterable which yields a single item
        return iterable,  # <- that comma is important!


# Intended to be also collections.Iterable
class GenericSplit(StorageObject):
    """Iterator to chain several chunks.

    @ClassField _chunks list<storageobject>
    @ClassField split_brothers list<storageobject>
    @ClassField storage_location anything
    @ClassField _current_chunk_idx int
    @ClassField _last_chunk_idx int
    """

    @dclayMethod(chunks="list<storageobject>", storage_location="anything")
    def __init__(self, chunks, storage_location):
        """Build a LocalIterator through a list of chunks.

        :param chunks: Sequence of (iterable) chunks.
        """
        # If this is not being called remotely, better to coerce to list right now
        self._chunks = list(chunks)
        self.storage_location = storage_location
        self.split_brothers = list()
        self._current_chunk_idx = -1
        self._last_chunk_idx = -1

        # non persistent field:
        self._current_iter_chunk = None

    @dclayMethod(return_=bool, _local=True)
    def _go_next_chunk(self):
        """Advance to next chunk.

        Prepare the internal chunk iterator. This is typically called when the
        current chunk iterator has finished. Return True if everything is ok,
        False if there is no valid next chunk.

        :return bool: Whether the advance is a success
        """
        self._current_chunk_idx += 1

        if self._current_chunk_idx < self._last_chunk_idx:
            self._current_iter_chunk = iter(self._chunks[self._current_chunk_idx])
            return True
        else:
            return False

    @dclayMethod(_local=True)
    def __next__(self):
        try:
            return next(self._current_iter_chunk)
        except StopIteration:
            logger.info("Iterator %s passing the chunk_idx %d", self.getID(), self._current_chunk_idx)

        if self._go_next_chunk():
            # At this point, for non-degenerate collections, means that there
            # is a chunk ready --so this is should not result in deep stack calls
            return next(self)
        else:
            raise StopIteration

    # Backwards compatibility to Python 2 (to be deprecated)
    @dclayMethod(_local=True)
    def next(self):
        return self.__next__()

    @dclayMethod()
    def _remote_iteration_init(self):
        """This method is used for remote initialization.

        This registered method is used by the local-execution `__iter__`. Given
        that that method is _local, a certain remote initialization must be
        ensured in order for the iteration to work smoothly.
        """
        self._current_chunk_idx = 0 if self._chunks else -1
        self._last_chunk_idx = len(self._chunks)

    # Note that the return is not serializable, thus the _local flag
    @dclayMethod(return_="anything", _local=True)
    def __iter__(self):
        # Prepare the remote stuff
        self._remote_iteration_init()

        # Prepare the iterators (non-serializable, `_local` specific)
        if self._chunks:
            self._current_iter_chunk = iter(self._chunks[0])
        else:
            self._current_iter_chunk = iter(list())

        # Now this object can be iterated
        return self


class WorkStealingSplit(GenericSplit):
    """Iterator to chain several chunks with simple Work Stealing addendum.

    The Work Stealing performed by this split is a simple one in which the
    chunk is returned to the thief, but no modifications are done (no movement,
    no reorganization of objects).

    @ClassField _chunks list<storageobject>
    @ClassField split_brothers list<storageobject>
    @ClassField storage_location anything
    @ClassField _current_chunk_idx int
    @ClassField _last_chunk_idx int
    """

    @dclayMethod(stolen_object="storageobject", return_="storageobject")
    def _post_stealing(self, stolen_object):
        """Not used by the simple base class WorkStealingSplit."""
        return stolen_object

    @dclayMethod(return_="storageobject")
    def steal_me(self):
        if self._current_chunk_idx < self._last_chunk_idx - 2:
            self._last_chunk_idx -= 1
            return self._chunks[self._last_chunk_idx]
        else:
            return None

    @dclayMethod(_local=True)
    def __next__(self):
        # TODO: the following lines (until "# Proceed to worksteal!") should be substituted by a super call
        try:
            return next(self._current_iter_chunk)
        except StopIteration:
            logger.info("Iterator %s passing the chunk_idx %d", self.getID(), self._current_chunk_idx)

        if self._go_next_chunk():
            return next(self)
        else:
            # Proceed to worksteal!
            import random
            # Use the list of brother locally [**]
            brothers = self.split_brothers
            while brothers:
                logger.info("Trying a brother...")
                victim_idx = random.randint(0, len(brothers) - 1)
                steal = brothers[victim_idx].steal_me()
                if steal is None:
                    logger.info("Split method could not steal from brother, removing")
                    brothers.pop(victim_idx)
                else:
                    logger.info("received %r", steal)
                    chunk_to_use = self._post_stealing(steal)
                    logger.info("using %r", chunk_to_use)
                    self._current_iter_chunk = iter(chunk_to_use)

                    # This will jump over the else
                    break
            else:
                # No valid targets to steal
                logger.info("No valid targets to steal")
                raise StopIteration

            # [**] re-store it for future usage (now that it may have been trimmed
            self.split_brothers = brothers
            return next(self)


class WorkMovingSplit(WorkStealingSplit):
    """Iterator to chain several chunks with Work Stealing through movement.

    The balancing performed by this split is similar to the simple Work Stealing
    but in this scenario the chunks are reorganized through movement. When there
    is a steal, the chunk is "physically" moved to the receiving end.

    @ClassField _chunks list<storageobject>
    @ClassField split_brothers list<storageobject>
    @ClassField storage_location anything
    @ClassField _current_chunk_idx int
    @ClassField _last_chunk_idx int
    """

    @dclayMethod(stolen_object="storageobject", return_="storageobject")
    def _post_stealing(self, stolen_object):
        """Once an object has been stolen, perform the movement."""
        # FIXME: We should not assume that this is server-side
        from dataclay.commonruntime.Runtime import getRuntime 
        from dataclay.commonruntime.Settings import settings
        getRuntime().move_object(stolen_object, settings.storage_id)
        return stolen_object


class SplittableCollectionMixin(object):
    """Mixin to help the model programmer.

    This mixin is intended to be use with Collections that have "chunks" (or
    some kind of internal partitioning) and desire to use them to provide high
    level "split iteration" abstractions.

    To provide support for SplittableCollections, include a get_chunks method
    which must return the list of chunks. Note that each chunk must have an
    iteration mechanism, with an structure that should resemble the original
    collection.
    """

    @dclayMethod(return_="anything")
    def get_chunks(self):
        try:
            return self.chunks
        except AttributeError:
            raise NotImplementedError("ChunkedCollections must either implement the get_chunks method "
                                      "or contain a `chunks` attribute.")

    # TODO: define the actual split method parameters (contact with Hecuba?).
    # TODO: Right now you can see "hardcoded" the "LocalIteration" behaviour
    @dclayMethod(return_="list<storageobject>", split_class="anything",
                 _local=True)
    def split(self, split_class=None):
        # TODO: this could be improved, library could cache stuff, or a joint call could be added
        from dataclay.commonruntime.Runtime import getRuntime
        location_chunks = sorted(
            (
                # Overly complex way to get a single ExecutionEnvironmentID
                # works in both Python2 & Python3
                next(iter(((getRuntime().get_all_locations(ch.get_object_id())).keys()))),
                ch
            ) for ch in self.get_chunks()
        )

        result = list()

        # FIXME: once support for imports has been implemented and tested, clean this up
        from itertools import groupby
        from operator import itemgetter
        from dataclay.contrib import splitting

        if split_class is None:
            split_class = splitting.GenericSplit
        elif isinstance(split_class, basestring):
            split_class = getattr(splitting, split_class)
        else:
            raise NotImplementedError("I could not understand %s (of type %s)" % 
                                      (split_class, type(split_class)))

        unused_exec_envs = set(getRuntime().get_execution_environments_info().keys())

        # FIXME: **********************************************************************************
        # FIXME: not using real split_class due to registration issues, but should be done that way
        # FIXME: **********************************************************************************

        for loc, chunks in groupby(location_chunks, key=itemgetter(0)):
            # Previous code using split intelligently:
            # split_object = split_class(map(itemgetter(1), chunks), loc)
            # split_object.make_persistent(backend_id=loc)
            # unused_exec_envs.remove(loc)
            # result.append(split_object)

            # Hacked around code:
            result.append(map(itemgetter(1), chunks))

        # Remove this
        return result

        # FIXME This is useful for WorkStealingSplit, but not for the others...
        for ee in unused_exec_envs:
            split_object = split_class(list(), ee)
            split_object.make_persistent(backend_id=ee)
            result.append(split_object)

        for split_object in result:
            # Note that this result is implicitly copied given that split is persistent
            # (maybe do the copy explicit? even when it is useless?)
            split_object.split_brothers = result

        return result




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class DataContract(ManagementObject):
    _fields = ["dataClayID",
               "applicantsNames",
               "beginDate",
               "endDate",
               "publicAvailable",
               ]

    _internal_fields = ["providerAccountID",
                        "providerDataSetID",
                        "applicantsAccountsIDs",
                        ]
